import sail

import time, pathlib, os
import json, click, requests, shlex, subprocess
import fabric, paramiko
import jinja2

class SailException(click.ClickException):
	def show(self, file=None):
		click.echo(click.style('Error: ', fg='red') + self.format_message(), err=True)

_debug = False
def debug(set=None):
	'''Get or set debug mode'''
	global _debug

	if set is None:
		return _debug

	_debug = set
	return _debug

def dlog(line):
	'''Print debug output'''
	if not debug():
		return

	click.secho('(debug) %s' % line, fg='yellow')

def request(endpoint, **kwargs):
	'''Request the Sail API endpoint'''

	method = kwargs.get('method')
	params = kwargs.get('params', {})
	json = kwargs.get('json', {})
	anon = kwargs.get('anon', False)

	api_base = get_sail_default('api-base')
	if not api_base:
		api_base = sail.API_BASE

	headers = {}

	if not anon:
		_config = config()
		headers = {
			'X-App-Id': _config['app_id'],
			'X-App-Secret': _config['secret'],
		}

	if not method:
		method = 'POST' if json else 'GET'

	url = '%s/%s' % (api_base.rstrip('/'), endpoint.lstrip('/'))

	if debug():
		safe_headers = headers.copy()
		if 'X-App-Secret' in safe_headers:
			safe_headers['X-App-Secret'] = '*'*8

		dlog('Requesting: [%s] %s' % (method, url))
		dlog('Request headers: %s' % repr(safe_headers))
		dlog('Request params: %s' % repr(params))
		dlog('Request json: %s' % repr(json))

	request = requests.Request(method, url, headers=headers, params=params)

	if method == 'POST' or json:
		request.json = json

	request = request.prepare()
	session = requests.Session()

	try:
		response = session.send(request)
	except Exception as e:
		dlog('Exception: %s' % repr(e))
		raise SailException('Could not contact the Sail API. Try again later.')

	dlog('Response: [%d] %s' % (response.status_code, repr(response.text)))

	try:
		data = response.json()
	except Exception as e:
		dlog('Exception: %s' % repr(e))
		raise SailException('Invalid response from the Sail API. Try again later.')

	if not response.ok:
		raise SailException('API error: %s' % data.get('error'))

	return data

def wait_for_task(task_id, timeout=30, interval=1):
	data = {}

	def _wait():
		nonlocal data

		data = request('/status/%s/' % task_id)
		if data.get('task_state') == 'failure':
			raise SailException('Task state failure')

		return data.get('task_state') == 'success'

	wait(_wait, timeout=timeout, interval=interval)
	return data

def wait(condition, timeout=30, interval=1, *args):
	'''Wait for condition every interval or timeout.'''
	start = time.time()
	while not condition(*args):
		if time.time() - start > timeout:
			raise Exception('Timeout in util.wait()')

		time.sleep(interval)

def find_root():
	p = pathlib.Path()
	p = p.resolve()

	while p.parent and str(p) != '/':
		for x in p.iterdir():
			if x.stem == '.sail':
				return str(p)

		p = p.parent

def update_config(data):
	root = find_root()
	with open(root + '/.sail/config.json', 'w+') as f:
		json.dump(data, f, indent='\t')

def config():
	root = find_root()

	if not root:
		raise SailException('Could not parse .sail/config.json. If this is a new project run: sail init')

	with open(root + '/.sail/config.json') as f:
		_config = json.load(f)

	if not _config:
		raise SailException('Could not parse .sail/config.json. If this is a new project run: sail init')

	# Back-compat
	if 'hostname' not in _config:
		_config['hostname'] = '%s.sailed.io' % _config['app_id']

	if 'namespace' not in _config:
		_config['namespace'] = 'default'

	if 'ssh_key_meta' not in _config:
		_config['ssh_key_meta'] = {}

	return _config

loader_i = 0
loader_suspended = False

def loader(suspend=None):
	global loader_i
	global loader_suspended

	if suspend:
		loader_suspended = True
		return

	if loader_suspended:
		return

	frames = [
		"   ",
		".  ",
		".. ",
		"...",
	]

	click.echo(frames[loader_i % len(frames)] + '\r', nl=False)
	loader_i += 1
	time.sleep(.2)

def get_sail_default(name):
	'''Get a sail default config variable'''
	import pathlib, json
	filename = (pathlib.Path.home() / '.sail-defaults.json').resolve()
	data = {}

	try:
		with open(filename) as f:
			data = json.load(f)
			return data[name]
	except:
		return None

def rsync(args, source, destination, default_filters=True, extend_filters=[]):
	root = find_root()
	args = args[:]

	if debug():
		args.extend(['-v'])

	filters = []

	if type(default_filters) != bool:
		raise Exception('default_filters expected to be bool')

	if default_filters:
		filters = [
			'- /wp-content/debug.log',
			'- /wp-content/uploads',
			'- /wp-content/cache',
			'- /wp-content/upgrade',
		]

	if extend_filters:
		filters.extend(extend_filters)

	# Force exclude all dot-files
	filters.insert(0, '- .*')

	args.insert(0, 'rsync')
	ssh_args = ['ssh',
		'-i', '%s/.sail/ssh.key' % root,
		'-o', 'UserKnownHostsFile="%s/.sail/known_hosts"' % root,
		'-o', 'IdentitiesOnly=yes',
		'-o', 'IdentityFile="%s/.sail/ssh.key"' % root,
	]

	args.extend(['-e', join(ssh_args)])

	# Add all filters in order
	for filter in filters:
		args.extend(['--filter', filter])

	args.extend([source, destination])

	if debug():
		dlog('Rsync: %s' % repr(args))

	# Download files FROM production
	p = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE, encoding='utf8')

	while True:
		try:
			stdout, stderr = p.communicate(timeout=0.2)
			break
		except subprocess.TimeoutExpired:
			loader()

	if debug():
		dlog('Rsync stdout: %s' % stdout)
		dlog('Rsync stderr: %s' % stderr)

	if 'WARNING: UNPROTECTED PRIVATE KEY FILE' in stderr:
		raise SailException('Unprotected private key file. Please fix permissions and try again.')

	return (p.returncode, stdout, stderr)

def connection():
	_config = config()
	root = find_root()

	ip = _config['ip']
	with open('%s/.sail/ssh.key' % root, 'r') as f:
		pkey = paramiko.RSAKey.from_private_key(f)

	ssh_config = fabric.Config()
	ssh_config.user = 'root'
	ssh_config.connect_kwargs = {'pkey': pkey, 'look_for_keys': False}
	ssh_config.run = {'echo': False}
	ssh_config.load_ssh_configs = False
	ssh_config.forward_agent = False

	c = fabric.Connection(ip, config=ssh_config)

	# Load known_hosts if it exists
	known_hosts = pathlib.Path(root) / '.sail/known_hosts'
	if known_hosts.exists() and known_hosts.is_file():
		c.client.load_host_keys(known_hosts)

	def _run(func):
		def run(*args, **kwargs):
			dlog('Fabric: %s, %s' % (repr(args), repr(kwargs)))
			kwargs['hide'] = True

			# Run it
			r = func(*args, **kwargs)

			stdout = r.stdout.strip()
			if stdout:
				dlog('Fabric stdout: %s' % stdout)

			stderr = r.stderr.strip()
			if stderr:
				dlog('Fabric stderr: %s' % stderr)

			return r
		return run

	# Decorate it
	c.run = _run(c.run)
	return c

def template(filename, data):
	e = jinja2.Environment(loader=jinja2.FileSystemLoader(sail.TEMPLATES_PATH))
	template = e.get_template(filename)
	return template.render(data)

def primary_url():
	_config = config()
	primary = [d for d in _config['domains'] if d['primary']][0]
	return ('https://' if primary.get('https') else 'http://') + primary['name']

def remote_path(directory=None):
	_config = config()
	namespace = _config.get('namespace', 'default')
	prefix = ''

	if namespace != 'default':
		prefix = '/_%s' % namespace

	path = '/var/www' + prefix
	if directory:
		path = path + '/' + directory.lstrip('/')

	return path

def join(split_command):
	return ' '.join(shlex.quote(arg) for arg in split_command)

def premium():
	_config = config()
	return bool(_config.get('premium', False))

def sizeof_fmt(num, suffix="B"):
	for unit in ['', 'Ki', 'Mi', 'Gi', 'Ti', 'Pi', 'Ei', 'Zi']:
		if abs(num) < 1024.0:
			return f'{num:3.1f}{unit}{suffix}'
		num /= 1024.0
	return f'{num:.1f}Yi{suffix}'

## Formatting:
is_silent = False
def silent(make_silent=None):
	global is_silent
	if make_silent is not None:
		is_silent = make_silent
		loader(suspend=is_silent)

	return is_silent

label_w = 0
def label_width(width):
	global label_w
	label_w = width

def label(label, fg='green'):
	global label_w
	return click.style(label.rjust(label_w), fg=fg)

def success(message, padding=True):
	if silent(): return
	if padding: click.echo()
	click.echo(click.style('Success: ', fg='green') + message)
	if padding: click.echo()

def failure(message):
	if silent(): return
	click.echo(click.style('Error: ', fg='red') + message)

def heading(message, padding=True):
	if silent(): return
	if padding: click.echo()
	click.echo(message)

def item(message):
	if silent(): return
	click.secho('  ' + message, fg='bright_black')



import sail

from sail import cli, util

import json, os
import click, pathlib

from datetime import datetime

@cli.group(invoke_without_command=True)
@click.pass_context
def backup(ctx):
	'''Create, restore and manage remote and local application backups'''
	# Default subcommand for back-compat
	if not ctx.invoked_subcommand:
		return ctx.forward(create)

@cli.command(name='restore')
@click.argument('path', nargs=1, required=True)
@click.option('--yes', '-y', is_flag=True, help='Skip the AYS message and force yes')
@click.option('--skip-db', is_flag=True, help='Do not import the database')
@click.option('--skip-uploads', is_flag=True, help='Do not import uploads')
@click.option('--nowait', is_flag=True, help='Do not wait for the remote task to complete and exit early')
@click.pass_context
def restore_compat(ctx, path, yes, skip_db, skip_uploads, nowait):
	'''Restore your application files, uploads and database from a local or remote backup'''
	return ctx.forward(restore)

@backup.command()
@click.argument('path', nargs=1, required=True)
@click.option('--yes', '-y', is_flag=True, help='Skip the AYS message and force yes')
@click.option('--skip-db', is_flag=True, help='Do not import the database')
@click.option('--skip-uploads', is_flag=True, help='Do not import uploads')
@click.option('--nowait', is_flag=True, help='Do not wait for the remote task to complete and exit early')
@click.pass_context
def restore(ctx, path, yes, skip_db, skip_uploads, nowait):
	'''Restore your application files, uploads and database from a local or remote backup'''
	config = util.config()

	if not path.isnumeric() or pathlib.Path(path).exists():
		return ctx.forward(sail.backups.restore)

	util.heading('Restoring backup')
	util.item('Requesting remote backup restore')

	request = util.request('/premium/backups/restore/', method='POST', json={
		'timestamp': path,
		'skip_db': skip_db,
		'skip_uploads': skip_uploads,
	})

	task_id = request['task_id']

	if nowait:
		util.success('Request received, exiting (nowait)')
		return

	util.item('Request received, waiting for task to complete')
	data = util.wait_for_task(task_id, 3600, 10)

	util.item('Task completed successfully')
	util.success('Remote backup restored. Local copy may be out of date!')

@backup.command()
@click.option('--local', is_flag=True, help='Force a local backup instead of remote')
@click.option('--description', help='Provide an optional description for this backup')
@click.option('--nowait', is_flag=True, help='Do not wait for the remote task to complete and exit early')
@click.pass_context
def create(ctx, local, description, nowait):
	'''Backup your production files and database to a remote or local location.'''
	config = util.config()

	if local and description:
		raise util.SailException('Descriptions are only supported for remote backups.')

	# Invoke the local backup command if asked for --local
	if local:
		return ctx.invoke(sail.backups.create)

	util.heading('Creating a backup')
	util.item('Requesting remote backup create')

	request = util.request('/premium/backups/create/', method='POST', json={
		'description': description,
	})

	task_id = request['task_id']

	if nowait:
		util.success('Request received, exiting (nowait)')
		return

	util.item('Request received, waiting for task to complete')
	data = util.wait_for_task(task_id, 3600, 10)

	util.item('Task completed successfully')
	util.success('Remote backup created successfully')

@backup.command(name='list')
@click.option('--json', 'as_json', is_flag=True, help='Return results as a JSON object')
def list_cmd(as_json):
	'''List available managed backups'''
	if not util.premium():
		raise util.SailException('Premium features are not enabled for this application. Learn more: https://sailed.io/premium/')

	backups = util.request('/premium/backups/list/')

	if as_json:
		click.echo(json.dumps(backups))
		return

	width, height = os.get_terminal_size()
	i = 0

	for backup in backups:
		timestamp = int(backup['timestamp'])

		ts = timestamp
		description = backup['description']
		size = util.sizeof_fmt(int(backup['size']))
		date = datetime.fromtimestamp(timestamp)

		# Truncate description
		maxlength = width - len(f'{ts} {size} {date}') - 4
		if len(description) > maxlength:
			description = description[:maxlength-5] + '...'

		dots_n = len(f'{ts} {description}') + len(f'{size} {date}')
		dots_n = width - dots_n - 4

		ts = click.style(ts, fg='green')
		dots = click.style('-'*dots_n, fg='bright_black')
		size = click.style(size, fg='bright_black')
		date = click.style(date, fg='bright_black')

		if i % (height / 2) == 0:
			click.echo()
			left = ' Timestamp  Description'
			right = 'Size           Date/Time '.rjust(width - len(left))
			click.secho(left + right, fg='bright_black')
			click.echo()

		click.echo(f' {ts} {description} {dots} {size} {date}')
		i += 1

	util.label_width(11)

	click.echo()
	label = util.label('Info:')
	click.echo(f'{label} sail backup info <timestamp>')

	label = util.label('Restore:')
	click.echo(f'{label} sail backup restore <timestamp>')

	label = util.label('Export:')
	click.echo(f'{label} sail backup export <timestamp>')
	click.echo()

@backup.command()
@click.argument('timestamp', nargs=1, required=True)
@click.option('--nowait', is_flag=True, help='Do not wait for the remote task to complete and exit early')
@click.pass_context
def export(ctx, timestamp, nowait):
	'''Export a remote backup to a downloadable .tar.gz archive'''
	config = util.config()

	util.heading('Exporting backup')
	util.item('Requesting remote backup export')

	request = util.request('/premium/backups/export/', method='POST', json={
		'timestamp': timestamp,
	})

	task_id = request['task_id']

	if nowait:
		util.success('Request received, exiting (nowait)')
		return

	util.item('Request received, waiting for task to complete')
	data = util.wait_for_task(task_id, 3600, 10)

	util.item('Task completed successfully')
	ctx.invoke(info, timestamp=timestamp)

@backup.command()
@click.argument('timestamp', nargs=1, required=True)
@click.option('--json', 'as_json', is_flag=True, help='Return results as a JSON object')
def info(timestamp, as_json):
	'''Get information about a remote backup'''
	config = util.config()

	backup = util.request('/premium/backups/info/', method='POST', json={
		'timestamp': timestamp,
	})

	if as_json:
		click.echo(json.dumps(backup))
		return

	util.label_width(13)
	timestamp = int(backup['timestamp'])

	click.echo()

	label = util.label('Backup:')
	click.echo(f'{label} {timestamp}')

	label = util.label('Description:')
	description = backup['description']
	click.echo(f'{label} {description}')

	label = util.label('Status:')
	status = backup['status']
	click.echo(f'{label} {status}')

	label = util.label('Date/Time:')
	dt = datetime.fromtimestamp(timestamp)
	click.echo(f'{label} {dt}')

	label = util.label('Size:')
	size = util.sizeof_fmt(int(backup['size']))
	click.echo(f'{label} {size}')

	click.echo()
	label = util.label('Restore:')
	click.echo(f'{label} sail backup restore {timestamp}')

	if backup.get('export') == 'pending':
		label = util.label('Export:')
		click.echo(f'{label} pending')
	elif backup.get('export'):
		label = util.label('Exported:')
		export_url = backup['export']['url']
		click.echo(f'{label} {export_url}')

		label = util.label('Until:')
		export_expires = datetime.fromtimestamp(int(backup['export']['expires']))
		click.echo(f'{label} {export_expires}')
	else:
		label = util.label('Export:')
		click.echo(f'{label} sail backup export {timestamp}')

	click.echo()



from sail import cli, util

import click, pathlib, json
import os, shlex, subprocess, io
import webbrowser

@cli.command('config')
@click.argument('name', required=True, nargs=1)
@click.argument('value', required=False)
@click.option('--delete', is_flag=True)
def config(name, value=None, delete=False):
	'''Set reusable config variables'''
	valid_names = [
		'provider-token',
		'email',
		'api-base',
		'keep',
		'premium',
	]

	if name not in valid_names:
		raise util.SailException('Invalid config name: %s' % name)

	filename = (pathlib.Path.home() / '.sail-defaults.json').resolve()
	data = {}

	try:
		with open(filename) as f:
			data = json.load(f)
	except:
		pass

	if not value and not delete:
		if name not in data:
			raise util.SailException('The option is not set')

		click.echo(data[name])
		return

	if value and delete:
		raise util.SailException('The --delete flag does not expect an option value')

	if delete and name not in data:
		raise util.SailException('The option is not set, nothing to delete')

	if delete:
		del data[name]
		click.echo('Option %s deleted' % name)
	else:
		data[name] = value
		click.echo('Option %s set' % name)

	with open(filename, 'w+') as f:
		json.dump(data, f, indent='\t')

@cli.command(context_settings=dict(ignore_unknown_options=True))
@click.argument('command', nargs=-1)
def wp(command):
	'''Run a WP-CLI command on the production host'''
	root = util.find_root()
	config = util.config()

	command = util.join(command)

	os.execlp('ssh', 'ssh', '-tt',
		'-i', '%s/.sail/ssh.key' % root,
		'-o', 'UserKnownHostsFile="%s/.sail/known_hosts"' % root,
		'-o', 'IdentitiesOnly=yes',
		'-o', 'IdentityFile="%s/.sail/ssh.key"' % root,
		'-o', 'LogLevel=QUIET',
		'root@%s' % config['hostname'],
		'sudo -u www-data bash -c "cd %s; wp %s"' % (util.remote_path('/public'), command)
	)

@cli.command()
def admin():
	'''Open your default web browser to wp-admin. Logs in automatically if supported'''
	config = util.config()
	c = util.connection()

	util.heading('Logging in to wp-admin')

	primary = [d for d in config['domains'] if d['primary']]
	if len(primary) < 1:
		raise util.SailException('Could not find primary domain')

	primary = primary[0]
	url = ('https://' if primary.get('https') else 'http://') + primary['name']
	email = config['email']

	util.item('Running remote-login')
	command = ['wp', 'sail', 'remote-login', '--email=%s' % email]
	try:
		r = c.run('sudo -u www-data bash -c "cd %s; %s"' % (util.remote_path('/public'), util.join(command)))
		r = json.loads(r.stdout.strip())
		url = url + '/?_sail_remote_login=%s&id=%d' % (r['key'], r['id'])
	except:
		util.item('Remote login failed')
		url = url + '/wp-login.php'
		util.item('Login URL: %s' % url)

		if webbrowser.open(url):
			util.success('Browser window opened at wp-login.')
		else:
			util.item('Could not open browser window')
			util.item('Login URL: %s' % url)
			click.echo()

		return

	util.item('Remote login success')
	if webbrowser.open(url):
		util.success('Browser window opened at wp-admin.')
	else:
		util.item('Could not open browser window')
		util.item('Login URL (valid for 30s): %s' % url)
		click.echo()

@cli.command()
@click.option('--nginx', is_flag=True)
@click.option('--php', is_flag=True)
@click.option('--nginx-access', is_flag=True)
@click.option('--nginx-error', '--nginx-errors', is_flag=True)
@click.option('--php-error', '--php-errors', is_flag=True)
@click.option('--postfix', '--mail', is_flag=True)
@click.option('--mysql', '--mariadb', is_flag=True)
@click.option('--follow', '-f', is_flag=True)
@click.option('--lines', '-n', type=int)
def logs(nginx, php, nginx_access, nginx_error, php_error, postfix, mysql, follow, lines):
	'''Query and follow logs from the production server'''
	root = util.find_root()
	config = util.config()

	# Global settings
	settings = []
	width, height = os.get_terminal_size()

	if follow:
		settings.append('--follow')
	elif lines:
		settings.append('--lines %d' % lines)
	else:
		settings.append('--lines %d' % int(height - 1))

	# Journald settings
	if postfix:
		settings.append('-t postfix/qmgr')
		settings.append('-t postfix/pickup')
		settings.append('-t postfix/master')
		settings.append('-t postfix/smtp')
		settings.append('-t postfix/cleanup')
		settings.append('-t postfix/postsuper')
		settings.append('-t postfix/postqueue')

	if php_error or php:
		settings.append('-t php')

	journald_settings = ' '.join(settings)
	command = 'journalctl --no-hostname --directory=/var/log/journal %s' % journald_settings

	# Non-journald logs.
	if nginx_access or nginx:
		command = 'tail /var/log/nginx/access.log %s' % ' '.join(settings)
		if '--follow' not in settings:
			command += ' | less -S +G'

	if nginx_error:
		command = 'tail /var/log/nginx/error.log %s' % ' '.join(settings)
		if '--follow' not in settings:
			command += ' | less -S +G'

	if mysql:
		command = 'tail /var/log/mysql/mariadb-slow.log %s' % ' '.join(settings)
		if '--follow' not in settings:
			command += ' | less -S +G'

	os.execlp('ssh', 'ssh', '-tt',
		'-i', '%s/.sail/ssh.key' % root,
		'-o', 'UserKnownHostsFile="%s/.sail/known_hosts"' % root,
		'-o', 'IdentitiesOnly=yes',
		'-o', 'IdentityFile="%s/.sail/ssh.key"' % root,
		'-o', 'LogLevel=QUIET',
		'root@%s' % config['hostname'],
		command
	)

@cli.command('info')
@click.option('--json', 'as_json', is_flag=True, help='Output as a JSON string')
def info(as_json):
	'''Show current sail information'''
	config = util.config()

	util.label_width(12)

	labels = {
		'app_id': 'App ID',
		'namespace': 'Namespace',
		'hostname': 'Hostname',
		'ip': 'IP Address',
		'version': 'Version',
	}

	if as_json:
		data = {}
		for key in labels:
			data[key] = config[key]

		return click.echo(json.dumps(data))

	click.echo()

	for key, label in labels.items():
		label = util.label(f'{label}:')
		value = config[key]
		click.echo(f'{label} {value}')

	click.echo()



import sail

from sail import cli, util

import click

@cli.group()
def sftp():
	'''Manage SFTP access on this server'''
	pass

@sftp.command()
def enable():
	'''Enable SFTP access on this server'''
	config = util.config()
	c = util.connection()

	util.heading('Enabling SFTP access for www-data')

	util.item('Copying authorized_keys')
	c.run('mkdir -p /var/www/.ssh')
	c.run('cp /root/.ssh/authorized_keys /var/www/.ssh/authorized_keys')
	c.run('chown www-data. /var/www/.ssh/authorized_keys')
	c.run('chmod 0600 /var/www/.ssh/authorized_keys')

	util.item('Enabling shell')
	c.run('usermod -s /bin/bash www-data')

	click.echo()
	util.label_width(10)

	label = util.label('Hostname:')
	hostname = config['hostname']
	click.echo(f'{label} {hostname}')
	label = util.label('Username:')
	click.echo(f'{label} www-data')
	label = util.label('SSH Key:')
	click.echo(f'{label} .sail/ssh.key')
	label = util.label('Port:')
	click.echo(f'{label} 22')

	util.success('SFTP enabled successfully')

@sftp.command()
def disable():
	'''Disable SFTP access for the www-data user'''
	c = util.connection()

	util.heading('Disabling SFTP access for www-data')

	util.item('Deleting authorized_keys')
	c.run('rm -f /var/www/.ssh/authorized_keys')

	util.item('Disabling shell')
	c.run('usermod -s /sbin/nologin www-data')

	util.success('SFTP disabled successfully')



from sail import cli, util

import unittest
from unittest.mock import Mock, patch, call
from click.testing import CliRunner

_find_root = Mock(return_value='/path/to/project')
_util_config = Mock(return_value={
	"app_id": "foo",
	"secret": "bar",
	"hostname": "foobar.justsailed.io",
	"provider_token": "token",
	"droplet_id": 123,
	"key_id": 456,
	"version": "0.10.0",
})
_request = Mock()
_rmtree = Mock()
_droplet = Mock()
_sshkey = Mock()
_delete_dns_records = Mock()

@patch('sail.util.find_root', _find_root)
@patch('sail.util.config', _util_config)
@patch('sail.util.request', _request)
@patch('sail.domains._delete_dns_records', _delete_dns_records)
@patch('shutil.rmtree', _rmtree)
@patch('digitalocean.Droplet', _droplet)
@patch('digitalocean.SSHKey', _sshkey)
class TestDestroy(unittest.TestCase):
	def setUp(self):
		_find_root.reset_mock()
		_util_config.reset_mock()
		_request.reset_mock()
		_rmtree.reset_mock()
		_droplet.reset_mock()
		_sshkey.reset_mock()

	def test_confirm(self):
		runner = CliRunner()
		result = runner.invoke(cli, ['destroy'])
		self.assertEqual(result.exit_code, 1)
		self.assertIn('Are you sure?', result.output)
		self.assertIn('Aborted!', result.output)
		_request.assert_not_called()
		_rmtree.assert_not_called()

	def test_destroy_input(self):
		runner = CliRunner()
		result = runner.invoke(cli, ['destroy'], input='y')
		self.assertEqual(result.exit_code, 0)
		_request.assert_called_once_with('/destroy/', method='DELETE')

		config = util.config()
		provider_token = config['provider_token']
		droplet_id = config['droplet_id']
		key_id = config['key_id']

		self.assertEqual(_droplet.mock_calls,
			[call(token=provider_token, id=droplet_id), call().destroy()])

		self.assertEqual(_sshkey.mock_calls,
			[call(token=provider_token, id=key_id), call().destroy()])

		self.assertIn('Droplet destroyed successfully', result.output)
		self.assertIn('SSH key destroyed successfully', result.output)

		_rmtree.assert_called_once_with('/path/to/project/.sail')

	def test_destroy_y(self):
		runner = CliRunner()
		result = runner.invoke(cli, ['destroy', '-y'])
		self.assertEqual(result.exit_code, 0)
		_request.assert_called_once_with('/destroy/', method='DELETE')

		config = util.config()
		provider_token = config['provider_token']
		droplet_id = config['droplet_id']
		key_id = config['key_id']

		self.assertEqual(_droplet.mock_calls,
			[call(token=provider_token, id=droplet_id), call().destroy()])

		self.assertEqual(_sshkey.mock_calls,
			[call(token=provider_token, id=key_id), call().destroy()])

		self.assertIn('Droplet destroyed successfully', result.output)
		self.assertIn('SSH key destroyed successfully', result.output)

		_rmtree.assert_called_once_with('/path/to/project/.sail')

	def test_help(self):
		runner = CliRunner()
		result = runner.invoke(cli, ['destroy', '--help'])
		self.assertEqual(result.exit_code, 0)
		self.assertIn('Usage:', result.output)
		_request.assert_not_called()
		_rmtree.assert_not_called()



import sail

from sail import cli, util, ssh

from . import monitor

from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives import serialization

import click, os

def premium_init():
	'''Initialize premium modules if in a premium-enabled application'''
	try:
		is_premium = util.premium()
		if not is_premium:
			raise Exception('Not premium, nothing to init.')

		_premium_force_init()
	except:
		pass

def _premium_force_init():
	from sail.premium import backups, monitor

@cli.group()
def premium():
	'''Enable or disable Sail Premium for your application'''
	pass

@premium.command()
@click.option('--force', is_flag=True, help='Force enable routine, even if already premium.')
@click.option('--email', help='The admin e-mail address. You can set the default e-mail address with: sail config email <email>')
@click.option('--license', help='The premium license key. You can set a default key with: sail config premium <license>')
@click.pass_context
def enable(ctx, force, email, license, doing_init=False):
	'''Provision premium features for this application.'''
	config = util.config()
	root = util.find_root()

	if not license:
		license = util.get_sail_default('premium')

	if not email:
		email = util.get_sail_default('email')

	util.heading('Setting up Sail Premium')

	if config.get('premium') and not force:
		raise util.SailException('Premium features have already been enabled for this application.')

	if not license:
		raise util.SailException('Premium license key not found. Set one with: sail config premium LICENSE_KEY')

	if not email:
		raise util.SailException('Premium requires an e-mail configuration. Set one with: sail config email EMAIL_ADDRESS')

	util.item('Verifying license key')
	response = util.request('/premium/check/', json={
		'email': email,
		'license': license,
	})

	if not response:
		raise util.SailException('Could not verify the premium license key.')

	util.item('Generating SSH keys')

	# Generate a key pair.
	key = rsa.generate_private_key(public_exponent=65537, key_size=2048)
	private_key = key.private_bytes(
		serialization.Encoding.PEM,
		serialization.PrivateFormat.TraditionalOpenSSL,
		serialization.NoEncryption()
	).decode('utf8')

	public_key = key.public_key().public_bytes(
		serialization.Encoding.OpenSSH,
		serialization.PublicFormat.OpenSSH
	).decode('utf8')

	with open('%s/.sail/premium.key.pub' % root, 'w+') as f:
		f.write(public_key)
	os.chmod('%s/.sail/premium.key.pub' % root, 0o644)

	# Updates config
	ctx.invoke(ssh.add, path='%s/.sail/premium.key.pub' % root, label='.sail/premium.key', quiet=True)
	config = util.config() # Reload it after update

	util.item('Verifying connection')
	response = util.request('/premium/enable/', json={
		'email': email,
		'license': license,
		'private_key': private_key,
		'namespace': config['namespace'],
	})

	c = util.connection()

	util.item('Updating remote configuration files')

	# Updated main/shared configs for <= 0.10.2.
	c.put(sail.TEMPLATES_PATH + '/nginx.main.conf', '/etc/nginx/nginx.conf')
	c.put(sail.TEMPLATES_PATH + '/nginx.shared.conf', '/etc/nginx/conf.d/extras/sail.conf')
	c.put(sail.TEMPLATES_PATH + '/prepend.php', '/etc/sail/prepend.php')

	# Premium configs.
	c.put(sail.TEMPLATES_PATH + '/nginx.main.premium.conf', '/etc/nginx/conf.d/extras/sail.main.premium.conf')
	c.put(sail.TEMPLATES_PATH + '/nginx.premium.conf', '/etc/nginx/conf.d/extras/sail.premium.conf')
	c.put(sail.TEMPLATES_PATH + '/premium.php', '/etc/sail/premium.php')

	# Make sure updated configs are live.
	c.run('systemctl reload nginx')

	util.item('Updating .sail/config.json')
	config['premium'] = license
	util.update_config(config)

	try:
		util.item('Enabling monitoring')
		ctx.invoke(monitor.enable, quiet=True)
	except:
		util.item('Monitoring could not be enabled!')

	# Don't show success during provision/init.
	if not doing_init:
		util.success('Premium has been enabled successfully')



from sail import cli, util, deploy, domains, blueprints, cron, ssh, __version__

import sail
import json
import subprocess, os
import click
import shutil
import digitalocean
import re, secrets, string
import shlex, time, io, pathlib, jinja2
import packaging.version

from types import SimpleNamespace

from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives import serialization

@cli.command()
@click.option('--provider-token', help='Your DigitalOcean API token, must be read-write. You can set a default token with: sail config provider-token <token>')
@click.option('--size', default='s-1vcpu-1gb-intel', help='The Droplet size, defaults to s-1vcpu-1gb-intel. To get a full list of available sizes and pricing run: sail sizes')
@click.option('--region', default='ams3', help='The region to deploy to, defaults to ams3. For a full list of available regions run: sail regions')
@click.option('--email', help='The admin e-mail address. You can set the default e-mail address with: sail config email <email>')
@click.option('--namespace', default='default', help='The namespace to use for the new application')
@click.option('--environment', help='Initialize the application into an existing environment')
@click.option('--blueprint', 'blueprint', default='default.yaml', help='Apply a blueprint after init, defaults to: default.yaml')
@click.option('--json', 'as_json', is_flag=True, help='Suspend regular output and print results in JSON format')
@click.option('--force', '-f', is_flag=True)
@click.pass_context
def init(ctx, provider_token, email, size, region, force, namespace, environment, blueprint, as_json):
	'''Initialize and provision a new project'''
	root = util.find_root()

	if as_json:
		util.silent(True)

	if root and os.path.exists(root + '/.sail'):
		raise util.SailException('This ship has already sailed. Pick another one or remove the .sail directory.')

	files = os.listdir(path='.')
	if files and not force:
		raise util.SailException('This project directory is not empty, can not init here. Override with --force, can destroy existing files.')

	if namespace and not re.match(r'^[a-zA-Z0-9_.-]+$', namespace):
		raise util.SailException('Invalid namespace. Namespaces can be alpha-numeric and contain the following characters: _.-')

	namespace = namespace.lower()

	if environment:
		environment = pathlib.Path(environment) / '.sail'
		if not environment.is_dir():
			raise util.SailException('Not a valid Sail environment: %s' % environment.resolve())

		if not (environment / 'config.json').is_file():
			raise util.SailException('Not a valid Sail environment: %s' % environment.resolve())

		if not (environment / 'ssh.key').is_file() or not (environment / 'ssh.key.pub').is_file():
			raise util.SailException('Could not read SSH keys from: %s' % environment.resolve())

		with (environment / 'config.json').open('r') as f:
			env_config = json.load(f)

		if packaging.version.parse(env_config.get('version')) < packaging.version.parse('0.10.0'):
			raise util.SailException('The target environment version is lower than the supported minimum. Please upgrade the environment first.')

		if env_config.get('namespace', 'default').lower() == namespace:
			raise util.SailException('The "%s" namespace is already in use in the target environment. Please use a different namespace.' % namespace)

		if not env_config.get('provider_token'):
			raise util.SailException('Could not find a provider token in the target environment.')

		# Force the same provider token
		provider_token = env_config.get('provider_token')

	if not provider_token:
		provider_token = util.get_sail_default('provider-token')

	if not provider_token:
		raise util.SailException('You need to provide a DigitalOcean API token with --provider-token, or set a default one with: sail config provider-token <token>')

	# Make sure the provider token works
	try:
		account = digitalocean.Account(token=provider_token)
		account.load()
		if account.status != 'active':
			raise util.SailException('This DigitalOcean account is not active.')
		if not account.email_verified:
			raise util.SailException('This DigitalOcean account e-mail is not verified.')
	except:
		raise util.SailException('Invalid provider token.')

	if not email:
		email = util.get_sail_default('email')

	if not email:
		raise util.SailException('You need to provide an admin e-mail address with --email, or set a default one with: sail config email <e-mail>')

	util.heading('Initializing')

	response = util.request('/init/', json={
		'email': email,
		'version': __version__,
	}, anon=True)

	app_id = response['app_id']
	secret = response['secret']
	hostname = response['hostname']

	util.item('Claimed application id: %s' % app_id)

	os.mkdir('.sail')
	root = util.find_root()

	util.item('Writing .sail/config.json')
	config = {
		'app_id': app_id,
		'secret': secret,
		'hostname': hostname,
		'provider_token': provider_token,
		'email': email,
		'domains': [],
		'profile_key': ''.join(secrets.choice(string.ascii_letters + string.digits) for i in range(32)),
		'namespace': namespace,
		'version': __version__,
	}

	if environment:
		config['profile_key'] = env_config['profile_key']
		config['ip'] = env_config['ip']
		config['droplet_id'] = env_config['droplet_id']
		config['key_id'] = env_config['key_id']

	util.update_config(config)

	passwords = {}
	for key in ['mysql', 'wp']:
		passwords[key] = ''.join(secrets.choice(string.ascii_letters
			+ string.digits) for i in range(48))

	# Provision a new environment or copy existing
	if not environment:
		_provision(provider_token, size, region)
		_configure()
	else:
		_copy_environment(environment)

	# Run the installs
	_install(passwords)

	c = util.connection()

	# Add the default WP cron schedule.
	ctx.invoke(cron.add, schedule='*/5', command=('wp cron event run --due-now',), quiet=True)

	# Run a default blueprint
	if blueprint and blueprint != 'no' and blueprint != 'none':
		try:
			ctx.invoke(blueprints.blueprint, path=[blueprint], doing_init=True)
			remote_path = util.remote_path()
			c.run(f'sudo -u www-data wp --path={remote_path}/public option get home')
		except:
			util.item('Could not apply blueprint')

	# Download files from production
	ctx.invoke(deploy.download, yes=True, doing_init=True)

	# Create a local empty wp-contents/upload directory
	content_dir = '%s/wp-content/uploads' % root
	if not os.path.exists(content_dir):
		os.mkdir(content_dir)

	# Automatically enable premium if it's been set up
	premium_license = util.get_sail_default('premium')
	premium_email = util.get_sail_default('email')
	if premium_license and premium_email:
		try:
			ctx.invoke(sail.premium.enable, doing_init=True)
		except:
			util.item('Could not initialize Premium')

	util.heading('Initialization successful')

	if as_json:
		_config = util.config()
		_config.update({'password': passwords['wp']})
		click.echo(json.dumps(_config))
		return
	else:
		_success(passwords['wp'])

def _success(wp_password):
	config = util.config()
	primary_url = util.primary_url()
	email = config['email']
	hostname = config['hostname']

	util.item('The ship has sailed!')
	click.echo()

	util.label_width(10)

	label = util.label('URL:')
	click.echo(f'{label} {primary_url}')

	label = util.label('Login:')
	click.echo(f'{label} {primary_url}/wp-login.php')

	label = util.label('E-mail:')
	click.echo(f'{label} {email}')

	label = util.label('Password:')
	click.echo(f'{label} {wp_password}')

	click.echo()

	label = util.label('SSH Host:')
	hostname = config['hostname']
	click.echo(f'{label} {hostname}')

	label = util.label('SSH Port:')
	click.echo(f'{label} 22')

	label = util.label('Username:')
	click.echo(f'{label} root')

	label = util.label('SSH Key:')
	click.echo(f'{label} .sail/ssh.key')

	label = util.label('App root:')
	app_root = util.remote_path()
	click.echo(f'{label} {app_root}')

	click.echo()
	click.echo('To open an interactive shell run: sail ssh')
	click.echo('For support and documentation visit sailed.io')
	click.echo()

def _copy_environment(environment):
	root = util.find_root()
	config = util.config()

	util.item('Copying environment')

	shutil.copy(environment / 'ssh.key', '%s/.sail/ssh.key' % root)
	os.chmod('%s/.sail/ssh.key' % root, 0o600)
	shutil.copy(environment / 'ssh.key.pub', '%s/.sail/ssh.key.pub' % root)
	os.chmod('%s/.sail/ssh.key.pub' % root, 0o644)

	util.item('Requesting DNS record')
	response = util.request('/ip/', json={
		'ip': config['ip'],
	})

	util.item('Checking SSH connection')
	c = util.connection()

	try:
		c.run('hostname')
	except Exception as e:
		raise e
		raise util.SailException('Could not connect via SSH. Make sure the environment is healthy.')

	try:
		remote_config = json.loads(c.run('cat /etc/sail/config.json').stdout)
	except:
		raise util.SailException('Could not read /etc/sail/config.json')

	if not remote_config.get('namespaces'):
		raise util.SailException('Could not find any namespaces data in the remote config. This environment looks broken.')

	# Make sure our new namespace is unique
	if config['namespace'] in remote_config['namespaces']:
		raise util.SailException('The "%s" namespace already exists in this environment.' % config['namespace'])

	remote_config['namespaces'].append(config['namespace'])
	c.put(io.StringIO(json.dumps(remote_config)), '/etc/sail/config.json')

	util.item('Writing server keys to .sail/known_hosts')
	with open('%s/.sail/known_hosts' % root, 'w+') as f:
		r = subprocess.run(['ssh-keyscan', '-t', 'rsa,ecdsa,ed25519', '-H', config['ip'],
			config['hostname']], stdout=f, stderr=subprocess.DEVNULL)

def _provision(provider_token, size, region):
	root = util.find_root()
	config = util.config()

	util.item('Provisioning servers')

	# Generate a key pair.
	key = rsa.generate_private_key(public_exponent=65537, key_size=2048)
	private_key = key.private_bytes(
		serialization.Encoding.PEM,
		serialization.PrivateFormat.TraditionalOpenSSL,
		serialization.NoEncryption()
	).decode('utf8')

	public_key = key.public_key().public_bytes(
		serialization.Encoding.OpenSSH,
		serialization.PublicFormat.OpenSSH
	).decode('utf8')

	util.item('Writing SSH keys to .sail/ssh.key')
	with open('%s/.sail/ssh.key' % root, 'w+') as f:
		f.write(private_key)
	os.chmod('%s/.sail/ssh.key' % root, 0o600)

	with open('%s/.sail/ssh.key.pub' % root, 'w+') as f:
		f.write(public_key)
	os.chmod('%s/.sail/ssh.key.pub' % root, 0o644)

	util.item('Uploading SSH key to DigitalOcean')

	key = digitalocean.SSHKey(
		token=provider_token,
		name=config['app_id'],
		public_key=public_key
	)

	try:
		key.create()
		key.load()
	except Exception as e:
		print(e)
		raise util.SailException('Could not upload SSH key. Make sure token is RW.')

	util.item('Creating a new Droplet')

	with open(sail.TEMPLATES_PATH + '/cloud-config.yaml', 'r') as f:
		cloud_config = f.read()

	droplet = digitalocean.Droplet(
		token=provider_token,
		name=config['hostname'],
		region=region,
		image=sail.DEFAULT_IMAGE,
		size_slug=size,
		monitoring=True,
		ssh_keys=[key.id],
		backups=False,
		user_data=cloud_config
	)

	try:
		droplet.create()
	except Exception as e:
		raise util.SailException('Cloud not create new droplet. Please try again later.')

	util.item('Waiting for Droplet to boot')

	completed = False
	while not completed:
		time.sleep(3)
		actions = droplet.get_actions()
		for action in actions:
			action.load()
			if 'completed' == action.status:
				completed = True
				break

	util.item('Waiting for IP address')

	def wait_for_ip():
		droplet.load()
		if droplet.ip_address:
			return True
		return False

	util.wait(wait_for_ip, timeout=120, interval=10)

	util.item('Droplet up and running, requesting DNS record')
	response = util.request('/ip/', json={
		'ip': droplet.ip_address,
	})

	config['ip'] = droplet.ip_address
	config['droplet_id'] = droplet.id
	config['key_id'] = key.id

	# Add Sail's key metadata
	sail_fp = ssh._fingerprint(public_key)
	config['ssh_key_meta'][sail_fp] = {
		'created': int(time.time()),
		'label': '.sail/ssh.key',
		'protected': True,
	}

	util.update_config(config)

	util.item('Waiting for SSH')
	c = util.connection()

	def wait_for_ssh():
		try:
			c.run('hostname')
			return True
		except Exception as e:
			util.dlog(repr(e))
			return False

	util.wait(wait_for_ssh, timeout=120, interval=10)

	util.item('Writing server keys to .sail/known_hosts')
	with open('%s/.sail/known_hosts' % root, 'w+') as f:
		r = subprocess.run(['ssh-keyscan', '-t', 'rsa,ecdsa,ed25519', '-H', config['ip'],
			config['hostname']], stdout=f, stderr=subprocess.DEVNULL)

def _configure():
	config = util.config()
	c = util.connection()

	util.item('Configuring software')

	# Generate an /etc/sail/config.json
	config_json = {
		'app_id': config['app_id'],
		'profile_key': config['profile_key'],
		'namespaces': [config['namespace']],
	}

	c.run('mkdir /etc/sail')
	c.put(io.StringIO(json.dumps(config_json)), '/etc/sail/config.json')

	def wait_for_cloud_init():
		try:
			r = c.run('cloud-init status -l')
			return 'status: done' in r.stdout
		except:
			pass

		return False

	util.item('Waiting for cloud-init to complete')
	util.wait(wait_for_cloud_init, timeout=900, interval=10)

	c.run('mkdir -p /etc/nginx/conf.d/extras')
	c.put(sail.TEMPLATES_PATH + '/nginx.main.conf', '/etc/nginx/nginx.conf', preserve_mode=False)
	c.put(sail.TEMPLATES_PATH + '/nginx.shared.conf', '/etc/nginx/conf.d/extras/sail.conf', preserve_mode=False)
	c.put(sail.TEMPLATES_PATH + '/nginx.certbot.conf', '/etc/nginx/conf.d/extras/certbot.conf', preserve_mode=False)
	c.put(sail.TEMPLATES_PATH + '/prepend.php', '/etc/sail/prepend.php', preserve_mode=False)

	php_config = pathlib.Path(c.run('php -r "echo PHP_CONFIG_FILE_PATH;"').stdout).parent # /etc/php/8.1
	c.put(sail.TEMPLATES_PATH + '/php.ini', str(php_config / 'fpm/php.ini'), preserve_mode=False)
	c.put(sail.TEMPLATES_PATH + '/php.ini', str(php_config / 'cli/php.ini'), preserve_mode=False)

	c.put(sail.TEMPLATES_PATH + '/mariadb.cnf', '/etc/mysql/mariadb.conf.d/90-sail.cnf', preserve_mode=False)
	c.put(sail.TEMPLATES_PATH + '/logrotate.conf', '/etc/logrotate.d/sail', preserve_mode=False)

	# Certbot post-deploy hook
	c.run('mkdir -p /etc/letsencrypt/renewal-hooks/deploy/')
	c.put(sail.TEMPLATES_PATH + '/certbot.deploy.sh', '/etc/letsencrypt/renewal-hooks/deploy/sail.sh', preserve_mode=False)
	c.run('chmod +x /etc/letsencrypt/renewal-hooks/deploy/sail.sh')

	# Make sure certbot.conf is in action.
	c.run('systemctl reload nginx.service')

	# Make sure MariaDB config is active.
	c.run('systemctl restart mariadb.service')

	# Get xhprof
	extension_dir = c.run('php -r "echo ini_get(\\\"extension_dir\\\");"').stdout
	c.run('curl -L https://github.com/kovshenin/xhprof/releases/download/0.10.7-sail/xhprof-20210902.so.gz -o /tmp/xhprof.so.gz')
	c.run(f'gunzip /tmp/xhprof.so.gz && mv /tmp/xhprof.so {extension_dir}/xhprof.so')

	# Get dhparams
	c.run('curl -L https://raw.githubusercontent.com/certbot/certbot/4abd81e2186eddc67551d61a8260440bd177d18d/certbot/certbot/ssl-dhparams.pem -o /etc/nginx/conf.d/extras/ssl-dhparams.pem')

	# Permissions
	c.run('chown -R www-data. /var/www')

	# Make sure any sshd configs are applied
	c.run('systemctl restart ssh.service')

def _install(passwords):
	config = util.config()
	c = util.connection()

	# Request an SSL cert
	util.item('Requesting SSL certificate')

	retry = 3
	https = False
	while retry:
		retry -= 1
		try:
			c.run('certbot -n certonly --register-unsafely-without-email --agree-tos --standalone --http-01-port 8088 -d %s' % config['hostname'])
			https = True
			break
		except:
			if retry:
				util.item('Certbot failed, retrying')
				time.sleep(10)
			else:
				util.item('Certbot failed, skipping SSL')

	# Update the domains config.
	config['domains'].append({
		'name': config['hostname'],
		'internal': True,
		'https': https,
		'primary': True
	})
	util.update_config(config)

	# Update nginx configs with cert data.
	util.item('Generating nginx configuration')
	domains._update_nginx_config()

	# Prepare release directories
	remote_path = util.remote_path()
	c.run('mkdir -p %s/releases/1337' % remote_path)
	c.run('mkdir -p %s/uploads' % remote_path)
	c.run('mkdir -p %s/profiles' % remote_path)
	c.run('chown -R www-data. %s' % remote_path)

	# Create a MySQL database
	util.item('Setting up the MySQL database')

	c.run('mysql -e "CREATE DATABASE \\`wordpress_%s\\`;"' % config['namespace'])
	c.run('mysql -e "CREATE USER \\`wordpress_%s\\`@localhost IDENTIFIED BY \'%s\'"' % (config['namespace'], passwords['mysql']))
	c.run('mysql -e "GRANT ALL PRIVILEGES ON \\`wordpress_%s\\`.* TO \\`wordpress_%s\\`@localhost;"' % (config['namespace'], config['namespace']))

	util.item('Downloading and installing WordPress')
	wp = 'sudo -u www-data wp --path=%s/releases/1337 ' % remote_path
	admin_user = config['email'].split('@')[0]

	c.run(wp + 'core download')
	c.run(wp + util.join([
		'config', 'create',
		'--dbname=wordpress_%s' % config['namespace'],
		'--dbuser=wordpress_%s' % config['namespace'],
		'--dbpass=%s' % passwords['mysql']
	]))
	c.run(wp + util.join([
		'core', 'install',
		'--url=%s' % util.primary_url(),
		'--title=Sailed',
		'--admin_user=%s' % admin_user,
		'--admin_password=%s' % passwords['wp'],
		'--admin_email=%s' % config['email'],
		'--skip-email'
	]))
	c.run(wp + util.join([
		'rewrite', 'structure', '/%postname%/'
	]))

	# Disable standard wp-cron (will spawn via system cron).
	c.run(wp + util.join([
		'config', 'set', 'DISABLE_WP_CRON', 'true', '--raw'
	]))

	# Run any outstanding events (like the cron check).
	c.run(wp + util.join([
		'cron', 'event', 'run', '--due-now'
	]))

	# Do da deploy.
	util.item('Cleaning up')
	c.run('rm -rf %s/public' % remote_path)
	c.run('ln -sfn %s/releases/1337 %s/public' % (remote_path, remote_path))
	c.run('rm -rf %s/public/wp-content/uploads && ln -sfn %s/uploads %s/public/wp-content/uploads' % (
		remote_path, remote_path, remote_path))

	# Reload services
	c.run('systemctl reload nginx.service')

	php_config = pathlib.Path(c.run('php -r "echo PHP_CONFIG_FILE_PATH;"').stdout).parent # /etc/php/8.1
	php_version = php_config.name
	c.run(f'systemctl reload php{php_version}-fpm.service')

@cli.command()
@click.option('--yes', '-y', is_flag=True, help='Force yes on the are-you-sure prompt')
@click.option('--environment', is_flag=True, help='Force destroy environment, even if other namespaces exist')
@click.option('--skip-dns', is_flag=True, help='Do not attempt to delete DNS records for associated domains')
@click.option('--json', 'as_json', is_flag=True, help='Suspend regular output and print results in JSON format')
def destroy(yes, environment, skip_dns, as_json):
	'''Destroy an application namespace and/or the environment'''
	root = util.find_root()
	config = util.config()

	if as_json:
		util.silent(True)

	if not yes:
		click.confirm('All application data will be scrubbed and irretrievable. Are you sure?', abort=True)

	# Prevent destroying applications with added domains.
	user_domains = [d['name'] for d in config.get('domains', []) if not d.get('internal')]
	if len(user_domains) > 0:
		raise util.SailException('Remove the following domains first: %s' % ', '.join(user_domains))

	util.heading('Destroying application')

	namespaces = []
	try:
		c = util.connection()
		remote_config = json.loads(c.run('cat /etc/sail/config.json').stdout)
		namespaces = remote_config.get('namespaces', [])
	except:
		pass

	# It's the only namespace, destroy the environment
	if len(namespaces) < 1 or namespaces == [config['namespace']]:
		environment = True
	elif not environment and config['namespace'] == 'default':
		raise util.SailException('You asked to destroy the default namespace, but other namespaces exist in this environment. Use --environment to destroy the entire environment.')

	try:
		util.item('Releasing .justsailed.io subdomain')
		data = util.request('/destroy/', method='DELETE')
	except:
		pass

	if not skip_dns:
		util.item('Removing DNS records')
		_domains = [d['name'] for d in config.get('domains', []) if not d['internal']]
		_domains, _subdomains = domains._parse_domains(_domains)
		domains._delete_dns_records(_domains, _subdomains)

	if environment:
		util.item('Destroying environment')
		_destroy_environment()
	else:
		util.item('Destroying namespace')
		_destroy_namespace()

	util.item('Removing .sail/*')
	shutil.rmtree(root + '/.sail')

	util.success('Destroyed successfully')

	if as_json:
		click.echo(json.dumps({'success': True}))

def _destroy_namespace():
	config = util.config()
	c = util.connection()

	namespace = config['namespace']

	c.run('rm -rf /var/www/_%s' % namespace)
	c.run('rm -rf /etc/nginx/conf.d/%s.conf' % namespace)
	c.run('mysql -e "DROP DATABASE \\`wordpress_%s\\`;"' % namespace)
	c.run('mysql -e "DROP USER \\`wordpress_%s\\`@localhost"' % namespace)

	# Update the remote config.
	remote_config = json.loads(c.run('cat /etc/sail/config.json').stdout)
	remote_config['namespaces'].remove(config['namespace'])
	c.put(io.StringIO(json.dumps(remote_config)), '/etc/sail/config.json')

	# TODO: Maybe revoke/delete SSL certificates for this namespace

	c.run('systemctl reload nginx')

def _destroy_environment():
	config = util.config()
	try:
		droplet = digitalocean.Droplet(
			token=config['provider_token'],
			id=config['droplet_id']
		)
		droplet.destroy()
		util.item('Droplet destroyed successfully')
	except Exception as e:
		util.item('Error destroying droplet')

	try:
		key = digitalocean.SSHKey(
			token=config['provider_token'],
			id=config['key_id']
		)
		key.destroy()
		util.item('SSH key destroyed successfully')
	except:
		util.item('Error destroying SSH key')

@cli.command()
@click.option('--provider-token', help='Your DigitalOcean API token, must be read-write. You can set a default token with: sail config provider-token <token>')
@click.option('--json', 'as_json', is_flag=True)
def sizes(provider_token, as_json):
	'''Get available droplet sizes'''
	root = util.find_root()

	# Try and get a provider token from .sail/config.json
	if not provider_token and root:
		config = util.config()
		provider_token = config.get('provider_token')

	if not provider_token:
		provider_token = util.get_sail_default('provider-token')

	if not provider_token:
		raise util.SailException('You need to provide a DigitalOcean API token with --provider-token, or set a default one with: sail config provider-token <token>')

	sizes = []
	manager = digitalocean.Manager(token=provider_token)
	for size in manager.get_all_sizes():
		if size.available:
			sizes.append(SimpleNamespace(
				slug=size.slug,
				price=size.price_monthly,
				description=size.description,
				memory=size.memory,
				vcpus=size.vcpus,
				disk=size.disk,
			))

	if as_json:
		click.echo(json.dumps([s.__dict__ for s in sizes]))
		return

	width, height = os.get_terminal_size()
	slug_max_length = max([len(s.slug) for s in sizes])
	price_max_length = max([len(str(int(s.price))) for s in sizes]) + 1
	disk_max_length = max([len(str(int(s.disk))) for s in sizes]) + 1
	memory_max_length = max([len(str(int(s.memory / 1024))) for s in sizes]) + 1

	i = 0
	for size in sizes:
		slug = size.slug.rjust(slug_max_length)
		description = size.description
		memory = (str(int(size.memory / 1024)) + 'G').rjust(memory_max_length)
		disk = (str(size.disk) + 'G').rjust(disk_max_length)
		vcpus = str(size.vcpus) + 'C'
		price = ('$' + str(int(size.price))).rjust(price_max_length)

		dots_n = len(f'{slug} {description}') + len(f'{vcpus} {memory} {disk} {price}')
		dots_n = width - dots_n - 4
		dots = click.style('-'*dots_n, fg='bright_black')

		# Add styles
		slug = click.style(slug, fg='green')
		description = click.style(description)
		memory = click.style(memory, fg='bright_black')
		disk = click.style(disk, fg='bright_black')
		vcpus = click.style(vcpus, fg='bright_black')
		price = click.style(price, fg='bright_black')

		if i % (height / 2) == 0:
			click.echo()
			header = 'Slug'.rjust(slug_max_length + 1)
			header += ' Description'
			header += 'CPU  RAM  Disk Price '.rjust(width-len(header))
			click.secho(header, fg='bright_black')
			click.echo()

		click.echo(f' {slug} {description} {dots} {vcpus} {memory} {disk} {price}')
		i += 1

	click.echo()

@cli.command()
@click.option('--provider-token', help='Your DigitalOcean API token, must be read-write. You can set a default token with: sail config provider-token <token>')
@click.option('--json', 'as_json', is_flag=True)
def regions(provider_token, as_json):
	'''Get available deployment regions'''
	root = util.find_root()

	# Try and get a provider token from .sail/config.json
	if not provider_token and root:
		config = util.config()
		provider_token = config.get('provider_token')

	if not provider_token:
		provider_token = util.get_sail_default('provider-token')

	if not provider_token:
		raise util.SailException('You need to provide a DigitalOcean API token with --provider-token, or set a default one with: sail config provider-token <token>')

	manager = digitalocean.Manager(token=provider_token)
	regions = []
	for region in manager.get_all_regions():
		if region.available:
			regions.append(SimpleNamespace(
				name=region.name,
				slug=region.slug,
			))

	if as_json:
		click.echo(json.dumps([r.__dict__ for r in regions]))
		return

	click.echo()
	click.secho(' Slug  Region Name', fg='bright_black')
	click.echo()

	for region in regions:
		click.secho(f' {region.slug}', fg='green', nl=False)
		click.secho(f'  {region.name}')

	click.echo()



from sail import cli, util

import sail
import click
import digitalocean
import re
import shlex, io, os
import yaml
import pathlib
import json
import invoke

@click.argument('path', nargs=-1, required=True)
@cli.command(context_settings=dict(ignore_unknown_options=True))
def blueprint(path, doing_init=False):
	'''Run a blueprint file against your application'''
	config = util.config()

	_args = path
	arguments = []
	options = {}
	flags = []

	for arg in _args:
		if not arg.startswith('-'):
			arguments.append(arg)
			continue

		# option or flag
		if '=' in arg:
			k, v = arg.split('=')
			options[k] = v
		else:
			flags.append(arg)

	path = arguments[0] # TODO: Support multiple paths
	path = pathlib.Path(path)

	# Try Sail's internal library of BPs
	if not path.exists() and path.parent == pathlib.Path('.'):
		path = pathlib.Path(__file__).parent / 'blueprints' / path.name

	if not path.exists():
		raise util.SailException('File does not exist')

	if not path.name.endswith('.yml') and not path.name.endswith('.yaml'):
		raise util.SailException('Blueprint files must be .yml or .yaml')

	with path.open() as f:
		s = f.read()

	def _parse_variables(match):
		name = match.group(1).strip()
		if name in vars:
			return json.dumps(vars[name])
		return None

	# Load user variables and fill from command line arguments if possible.
	y = yaml.safe_load(s)
	vars = {}
	for var in y.get('vars', []):
		option = var.get('option')
		_type = str

		_map = {
			'str': str,
			'string': str,
			'int': int,
			'integer': int,
			'float': float,
			'bool': bool,
			'boolean': bool,
		}

		if var.get('type') and var.get('type') in _map.keys():
			_type = _map[var.get('type')]

		if options.get(option):
			value = options.get(option)
		else:
			value = click.prompt(var['prompt'], default=var.get('default', None), type=_type)

		if _type == bool and type(value) is not bool:
			truthy = ['yes', 'y', 'true', '1', 'affirmative']
			value = True if value.lower() in truthy else False

		elif _type == int or _type == float:
			try:
				value = _type(value)
			except ValueError:
				raise util.SailException('Could not convert %s to %s' % (repr(value), repr(_type)))

		vars[var['name']] = value

	# Reload with substitutions.
	s = re.sub(r'\${{([^}]+?)}}', _parse_variables, s)
	blueprint = yaml.safe_load(s)

	util.heading('Applying blueprint: %s' % path.name)

	for section, data in blueprint.items():
		if section == 'plugins' or section == 'themes':
			try:
				_bp_install_wp_products(section, data)
			except invoke.exceptions.UnexpectedExit as e:
				raise util.SailException(e.result.stderr)

		elif section == 'options':
			try:
				_bp_update_options(data)
			except invoke.exceptions.UnexpectedExit as e:
				raise util.SailException(e.result.stderr)
			continue

		elif section == 'define':
			try:
				_bp_define_constants(data)
			except invoke.exceptions.UnexpectedExit as e:
				raise util.SailException(e.result.stderr)

		elif section == 'fail2ban':
			_bp_fail2ban(data)

		elif section == 'postfix':
			_bp_postfix(data)

		elif section == 'dns':
			_bp_dns(data)

		elif section == 'apt':
			_bp_apt(data)

		elif section == 'files':
			_bp_files(data)

		elif section == 'commands':
			_bp_commands(data)

	# Don't show success during provision
	if not doing_init:
		util.success('Blueprint applied successfully')

def _bp_commands(data):
	c = util.connection()

	util.item('Running SSH commands')
	for command in data:
		if type(command) is list:
			command = util.join(command)

		command = 'cd %s && %s' % (util.remote_path('/public'), command)
		# command = 'sudo -u www-data bash -c "cd %s; bash"' % util.remote_path('/public')
		command = util.join(['sudo', '-u', 'www-data', 'bash', '-c', command])
		c.run(command, timeout=30)

def _bp_files(data):
	c = util.connection()

	util.item('Determining home directory')
	home = c.run('pwd').stdout.strip()
	home = pathlib.Path(home)

	for source, destination in data.items():
		p = pathlib.Path(os.path.expanduser(source))

		if not p.exists():
			util.item(f'Skipping: {source} (does not exist)')
			continue

		if type(destination) is str:
			destination = [destination]

		util.item(f'Copying {source}')
		for d in destination:
			d = d.replace('~', str(home))
			c.put(p, d)

def _bp_apt(items):
	c = util.connection()

	if 'selections' in items:
		util.item('Setting debconf selections')
		selections = items['selections']
		for line in selections:
			command = util.join(['echo', line]) + ' | debconf-set-selections'
			c.run(command)

	if 'install' in items:
		wait = 'while fuser /var/{lib/{dpkg,apt/lists},cache/apt/archives}/{lock,lock-frontend} >/dev/null 2>&1; do sleep 1; done && '
		util.item('Updating package lists')
		c.run(wait + 'DEBIAN_FRONTEND=noninteractive apt update', timeout=300)

		util.item('Installing packages')
		packages = items['install']
		command = wait + util.join(['DEBIAN_FRONTEND=noninteractive', 'apt', 'install', '-y'] + packages)

		try:
			c.run(command, timeout=300)
		except invoke.exceptions.UnexpectedExit as e:
			err = e.result.stderr
			err = err.replace('WARNING: apt does not have a stable CLI interface. Use with caution in scripts.', '')
			err = err.strip()
			raise util.SailException('Error: %s' % err)
		except invoke.exceptions.CommandTimedOut as e:
			raise util.SailException('Timeout')

def _bp_dns(records):
	config = util.config()

	# group records by domain
	domains = {}
	for record in records:
		if record['domain'] not in domains:
			domains[record['domain']] = []

		domains[record['domain']].append(record)

	# Make sure all domains exist
	for domain, records in domains.items():
		if domain not in [d['name'] for d in config['domains']]:
			raise util.SailException('This domain does not exist. To add use: sail domain add')

	for domain, records in domains.items():
		do_domain = digitalocean.Domain(token=config['provider_token'], name=domain)
		do_records = do_domain.get_records()

		for record in records:
			exists = False
			for do_record in do_records:
				if (do_record.name.lower() == record['name'].lower()
					and do_record.type.lower() == record['type'].lower()):
					exists = True
					continue

			if exists:
				util.item('Skipping %s record for %s.%s, already exists' %
					(record['type'], record['name'], record['domain']))
				continue

			util.item('Creating %s record for %s.%s' %
				(record['type'], record['name'], record['domain']))

			try:
				if record['type'].upper() == 'MX':
					do_domain.create_new_domain_record(
						name=record['name'],
						type=record['type'],
						data=record['value'],
						priority=record['priority']
					)
				else:
					do_domain.create_new_domain_record(
						name=record['name'],
						type=record['type'],
						data=record['value']
					)
			except Exception as e:
				raise util.SailException(str(e))

def _bp_postfix(data):
	config = util.config()
	namespace = config['namespace']
	c = util.connection()

	mode = data.get('mode')
	if mode != 'relay':
		raise util.SailException('Unsupported mode')

	relay_host = data.get('host')
	if not relay_host:
		raise util.SailException('Invalid relay host')

	wait = 'while fuser /var/{lib/{dpkg,apt/lists},cache/apt/archives}/{lock,lock-frontend} >/dev/null 2>&1; do sleep 1; done && '
	status = c.run('dpkg -s postfix', warn=True)
	if not status or status.stdout.find('Status: install ok installed') < 0:
		util.item('Installing postfix')
		c.run('debconf-set-selections <<< \'postfix postfix/main_mailer_type select Satellite system\'')
		c.run('debconf-set-selections <<< \'postfix postfix/mailname string %s\'' % config['hostname'])
		c.run('debconf-set-selections <<< \'postfix postfix/relayhost string %s\'' % relay_host)
		c.run(wait + 'apt update && DEBIAN_FRONTEND=noninteractive apt install -y postfix libsasl2-modules', timeout=300)
		c.run('usermod -a -G mail www-data', warn=True)
		c.run('usermod -a -G mail postfix', warn=True)

	util.item('Configuring postfix')

	try:
		postfix_config = json.loads(c.run('cat /etc/sail/postfix.json').stdout)
		util.item('Updating existing /etc/sail/postfix.json')
	except:
		util.item('Creating new /etc/sail/postfix.json')
		postfix_config = {}

	postfix_config[namespace] = {
		'namespace': namespace,
		'smtp_sender': data.get('from_email'),
		'smtp_host': data.get('host'),
		'smtp_port': data.get('port', 578),
		'smtp_username': data.get('username'),
		'smtp_password': data.get('password'),
	}

	c.put(io.StringIO(json.dumps(postfix_config)), '/etc/sail/postfix.json')
	c.run('chmod 0600 /etc/sail/postfix.json')

	sasl_passwd = []
	relay_hosts = []
	for entry in postfix_config.values():
		sasl_passwd.append(f"{entry['smtp_sender']} {entry['smtp_username']}:{entry['smtp_password']}")
		relay_hosts.append(f"{entry['smtp_sender']} {entry['smtp_host']}:{entry['smtp_port']}")

	sasl_passwd = '\n'.join(sasl_passwd)
	relay_hosts = '\n'.join(relay_hosts)

	c.put(io.StringIO(sasl_passwd), '/etc/postfix/sasl_passwd')
	c.put(io.StringIO(relay_hosts), '/etc/postfix/relay_hosts')
	c.run('chmod 0600 /etc/postfix/sasl_passwd')
	c.run('postmap /etc/postfix/sasl_passwd')
	c.run('postmap /etc/postfix/relay_hosts')

	c.put(io.StringIO(util.template('postfix/main.cf', {'hostname': config['hostname']})), '/etc/postfix/main.cf')
	c.run('chmod 0640 /etc/postfix/main.cf && chown root.mail /etc/postfix/main.cf')
	c.run('systemctl restart postfix.service')
	c.run('postfix set-permissions')

	# Set the from name/email in a mu-plugin
	if data.get('from_name') and data.get('from_email'):
		context = {
			'name': json.dumps(data.get('from_name')),
			'email': json.dumps(data.get('from_email')),
		}

		remote_path = util.remote_path()

		c.run('sudo -u www-data mkdir -p %s/public/wp-content/mu-plugins/' % remote_path)
		c.put(io.StringIO(util.template('postfix/1-sail-mail-from.php', context)),
			'%s/public/wp-content/mu-plugins/1-sail-mail-from.php' % remote_path)
		c.run('chown www-data. %s/public/wp-content/mu-plugins/1-sail-mail-from.php' % remote_path)

def _bp_fail2ban(jails):
	c = util.connection()
	remote_path = util.remote_path()

	wait = 'while fuser /var/{lib/{dpkg,apt/lists},cache/apt/archives}/{lock,lock-frontend} >/dev/null 2>&1; do sleep 1; done && '
	status = c.run('dpkg -s fail2ban', warn=True)
	if not status or status.stdout.find('Status: install ok installed') < 0:
		util.item('Installing fail2ban')
		c.run(wait + 'apt update && apt install -y fail2ban', timeout=300)

	util.item('Configuring fail2ban rules')
	# Make sure mu-plugins exists
	c.run('sudo -u www-data mkdir -p %s/public/wp-content/mu-plugins/' % remote_path)

	c.put(sail.TEMPLATES_PATH + '/fail2ban/wordpress-auth.conf', '/etc/fail2ban/filter.d/wordpress-auth.conf')
	c.put(sail.TEMPLATES_PATH + '/fail2ban/wordpress-pingback.conf', '/etc/fail2ban/filter.d/wordpress-pingback.conf')
	c.put(sail.TEMPLATES_PATH + '/fail2ban/nginx-deny.conf', '/etc/fail2ban/action.d/nginx-deny.conf')
	c.put(sail.TEMPLATES_PATH + '/fail2ban/0-sail-auth-syslog.php', '%s/public/wp-content/mu-plugins/0-sail-auth-syslog.php' % remote_path)
	c.put(io.StringIO(util.template('fail2ban/jail.local', {'jails': jails})), '/etc/fail2ban/jail.local')

	# Make sure configs directory exists and permissions ok
	c.run('chown www-data. %s/public/wp-content/mu-plugins/0-sail-auth-syslog.php' % remote_path)
	c.run('sudo -u www-data mkdir -p /var/www/configs')
	c.run('systemctl start fail2ban')
	c.run('fail2ban-client reload')

def _bp_define_constants(constants):
	c = util.connection()

	util.item('Updating wp-config.php constants')
	wp = 'sudo -u www-data wp --path=%s --skip-themes --skip-plugins ' % util.remote_path('/public')

	for name, value in constants.items():
		if type(value) not in [str, float, int, bool]:
			raise util.SailException('Invalid data type for %s' % name)

		raw = []

		if type(value) is bool:
			value = 'true' if value else 'false'
			raw = ['--raw']

		elif type(value) in [int, float]:
			value = str(value)
			raw = ['--raw']

		c.run(wp + util.join(['config', 'set', name, value] + raw), timeout=30)

def _bp_update_options(options):
	c = util.connection()

	util.item('Applying options')
	wp = 'sudo -u www-data wp --path=%s --skip-themes --skip-plugins ' % util.remote_path('/public')

	for option_name, data in options.items():
		# Scalars
		option_value = None
		format = []
		autoload = None
		delete = None

		if type(data) == dict:
			format = ['--format=json'] if data.get('type') == 'json' else []
			delete = data.get('delete', False)

			if data.get('autoload') is not None:
				autoload = bool(data.get('autoload'))

			option_value = data.get('value')
		else:
			option_value = data

		if not delete and type(option_value) not in [str, int, float]:
			raise util.SailException('Invalid value type for %s' % option_name)

		if type(option_value) is str:
			option_value = option_value.strip('\n')

		if delete:
			c.run(wp + util.join(['option', 'delete', option_name]), timeout=30)
		else:
			c.run(wp + util.join(['option', 'update', option_name,
				option_value] + format), timeout=30)

			# Set the autoload flag
			if autoload is not None:
				c.run(wp + util.join([
					'eval', "update_option( %s, get_option( %s ), %s );" %
					(json.dumps(option_name), json.dumps(option_name), 'true' if autoload else 'false')
				]))

def _bp_install_wp_products(what, products):
	wporg = {}
	custom = {}

	# what = plugins/themes
	for slug, data in products.items():
		if type(data) in [str, int, float]:
			wporg[slug] = data
			continue

		if type(data) == dict:
			if not data.get('url'):
				raise util.SailException('Could not find %s: %s' % (what[:-1], slug))

			custom[slug] = data
			continue

		raise util.SailException('Unknown %s specification: %s' % (what[:-1], slug))

	if not wporg and not custom:
		return

	c = util.connection()
	wp = 'sudo -u www-data wp --path=%s ' % util.remote_path('/public')

	util.item('Installing %s' % what)

	for slug, version in wporg.items():
		util.item('wporg/%s=%s' % (slug, version))
		_version = ['--version=%s' % version] if version != 'latest' else []
		r = c.run(wp + util.join([what[:-1], 'install', '--force', slug] + _version), timeout=30)

	for slug, data in custom.items():
		url = data.get('url')
		util.item('thirdparty/%s' % slug)
		c.run(wp + util.join([what[:-1], 'install', '--force', url]), timeout=30)

	if what == 'plugins':
		util.item('Activating plugins')

		if wporg:
			c.run(wp + util.join(['plugin', 'activate'] + list(wporg.keys())), timeout=60)

		if custom:
			c.run(wp + util.join(['plugin', 'activate'] + list(custom.keys())), timeout=60)

	else: # themes
		util.item('Activating theme')
		last = list(products.keys())[-1]
		c.run(wp + util.join(['theme', 'activate', last]), timeout=60)



from sail import cli, util

import subprocess, time
import click, pathlib
import re, shlex, os, stat

from glob import glob

def _get_extend_filters(paths, prefix=None):
	root = util.find_root()

	extend_filters = []
	if not paths:
		return []

	if prefix:
		prefixed_root = root.rstrip('/') + '/' + prefix.strip('/')

	for entry in paths:
		_entry = pathlib.Path(entry)

		# Requested root directory, remove all filters.
		if str(_entry.resolve()) == root:
			return []

		try:
			relative = _entry.resolve().relative_to(root)
		except ValueError:
			raise util.SailException('Could not resolve path: %s' % entry)

		if prefix:
			try:
				relative = _entry.resolve().relative_to(prefixed_root)
			except ValueError:
				continue # The item is in root, but not in relative root, skip

			if str(relative) == '.':
				continue

		for parent in reversed(relative.parents):
			if str(parent) == '.':
				continue
			extend_filters.append('+ /%s' % parent)

		if _entry.is_dir():
			extend_filters.append('+ /%s/***' % relative)
		else:
			extend_filters.append('+ /%s' % relative)

	# Skip everything else.
	if len(extend_filters) > 0:
		extend_filters.append('- *')

	return extend_filters

def _get_deployignore_filters():
	root = util.find_root()
	p = pathlib.Path(root) / '.deployignore'

	filters = []
	negates = []

	if not p.is_file():
		return filters

	with p.open('r') as f:
		lines = f.readlines()

	for line in lines:
		line = line.strip()
		if re.match(r'^\s*#', line):
			continue

		negate = re.match(r'^\s*!(.+)', line)
		if negate:
			negates.append('+ %s' % negate.group(1))
			continue

		filters.append('- %s' % line)

	# Negates have to come first for rsync.
	filters = negates + filters
	return filters

@cli.command()
@click.argument('path', nargs=-1, required=False)
@click.option('--with-uploads', is_flag=True, help='Include the wp-content/uploads directory')
@click.option('--dry-run', is_flag=True, help='Show changes about to be deployed to production')
@click.option('--skip-hooks', '--no-verify', is_flag=True, help='Do not run pre-deploy hooks')
@click.option('--redeploy', is_flag=True, help='Redeploy to an existing release directory')
@click.pass_context
def deploy(ctx, with_uploads, dry_run, path, skip_hooks, redeploy):
	'''Deploy your working copy to production. If path is not specified then all application files are deployed.'''
	root = util.find_root()
	config = util.config()

	release = str(int(time.time()))

	if dry_run:
		return ctx.invoke(diff, path=path)

	util.heading('Deploying to production')

	hooks = []
	failed = []

	if not skip_hooks:
		hooks = glob(root + '/.sail/pre-deploy') + glob(root + '/.sail/pre-deploy.*')

	if len(hooks) > 0:
		util.item('Running pre-deploy hooks')

	for hook in hooks:
		f = pathlib.Path(hook)
		if not os.access(f, os.X_OK):
			f.chmod(f.stat().st_mode | stat.S_IEXEC)

		p = subprocess.Popen(str(f), encoding='utf8', shell=True)
		p.communicate()
		if p.returncode > 0:
			failed.append(f)

	if len(failed) > 0:
		util.item('One or more pre-deploy hooks failed. Aborting deploy.')
		for f in failed:
			util.item('Failed: %s' % f)

		exit()

	c = util.connection()
	remote_path = util.remote_path()

	if redeploy:
		try:
			_current = c.run('readlink %s/public' % remote_path).stdout.strip().split('/')[-1]
			if not _current:
				raise Exception()
		except:
			raise util.SailException('Could not determine current release')

		util.item(f'Overwriting existing release: {_current}')
		release = _current

	else:
		util.item('Preparing release directory')
		c.run('mkdir -p %s/releases/%s' % (remote_path, release))
		c.run('rsync -rogtl %s/public/ %s/releases/%s' % (remote_path, remote_path, release))

	util.item('Uploading application files to production')

	# Parse .deployignore.
	filters = _get_deployignore_filters()
	filters += _get_extend_filters(path)

	returncode, stdout, stderr = util.rsync(
		args=['-rtl', '--rsync-path', 'sudo -u www-data rsync',
			'--copy-dest', '%s/public/' % remote_path, '--delete'],
		source='%s/' % root,
		destination='root@%s:%s/releases/%s' % (config['hostname'], remote_path, release),
		extend_filters=filters
	)

	if returncode != 0:
		raise util.SailException('An error occurred during upload. Please try again.')

	if with_uploads:
		util.item('Uploading wp-content/uploads')

		# Send uploads to production
		returncode, stdout, stderr = util.rsync(
			args=['-rtl', '--rsync-path', 'sudo -u www-data rsync', '--delete'],
			source='%s/wp-content/uploads/' % root,
			destination='root@%s:%s/uploads/' % (config['hostname'], remote_path),
			default_filters=False,
			extend_filters=_get_extend_filters(path, 'wp-content/uploads')
		)

		if returncode != 0:
			raise util.SailException('An error occurred during upload. Please try again.')

	util.item('Deploying release: %s' % release)

	if not redeploy:
		util.item('Updating symlinks')
		c.run('sudo -u www-data ln -sfn %s/uploads %s/releases/%s/wp-content/uploads' % (remote_path, remote_path, release))
		c.run('sudo -u www-data ln -sfn %s/releases/%s %s/public' % (remote_path, release, remote_path))

		util.item('Reloading services')
		c.run('nginx -s reload')

		php_config = pathlib.Path(c.run('php -r "echo PHP_CONFIG_FILE_PATH;"').stdout).parent # /etc/php/8.1
		php_version = php_config.name
		c.run(f'kill -s USR2 $(cat /var/run/php/php{php_version}-fpm.pid)')
	else:
		util.item('Nothing to update/reload in redeploy')

	releases = c.run('ls %s/releases' % remote_path)
	releases = re.findall('\d+', releases.stdout)
	releases = [int(i) for i in releases]

	keep = util.get_sail_default('keep')
	if not keep:
		keep = 5

	keep = int(keep)
	keep = max(keep, 2)
	keep = min(keep, 30)

	if len(releases) > keep:
		util.item('Removing outdated releases')
		remove = sorted(releases)[:len(releases)-keep]
		for key in remove:
			c.run(util.join(['rm', '-rf', '%s/releases/%s' % (remote_path, key)]))

	util.success('Successfully deployed %s' % release)

@cli.command()
@click.argument('release', required=False, type=int, nargs=1)
@click.option('--releases', is_flag=True, help='Get a list of valid releases to rollback to')
def rollback(release=None, releases=False):
	'''Rollback production to a previous release'''
	sail = util.config()
	c = util.connection()
	remote_path = util.remote_path()

	if releases or not release:
		util.heading('Fetching available releases')

		_releases = c.run('ls %s/releases' % remote_path)
		_releases = re.findall('\d+', _releases.stdout)

		if len(_releases) < 1:
			raise util.SailException('Could not find any releases')

		try:
			util.item('Determining current release')
			_current = c.run('readlink %s/public' % remote_path).stdout.strip().split('/')[-1]
		except:
			_current = '0'

		click.echo()

		for r in _releases:
			flags = '(current)' if r == _current else ''
			click.secho('  %s %s' % (r, flags))

		click.echo()
		click.echo('Rollback: sail rollback <release>')
		click.echo()
		return

	if release:
		release = str(release)

	util.heading('Rolling back to %s' % release)

	util.item('Fetching releases')
	_releases = c.run('ls %s/releases' % remote_path).stdout.strip().split('\n')
	if release not in _releases:
		raise util.SailException('Invalid release. To get a list run: sail rollback --releases')

	util.item('Updating symlinks')
	c.run('ln -sfn %s/releases/%s %s/public' % (remote_path, release, remote_path))

	util.item('Reloading services')
	c.run('nginx -s reload')

	php_config = pathlib.Path(c.run('php -r "echo PHP_CONFIG_FILE_PATH;"').stdout).parent # /etc/php/8.1
	php_version = php_config.name
	c.run(f'kill -s USR2 $(cat /var/run/php/php{php_version}-fpm.pid)')

	util.success('Successfully rolled back to %s' % release)

@cli.command()
@click.argument('path', nargs=-1, required=False)
@click.option('--yes', '-y', is_flag=True, help='Force Y on overwriting local copy')
@click.option('--with-uploads', is_flag=True, help='Include the wp-content/uploads directory')
@click.option('--delete', is_flag=True, help='Delete files from local copy that do not exist on production')
@click.option('--dry-run', is_flag=True, help='Show changes about to be downloaded to the working copy')
@click.pass_context
def download(ctx, path, yes, with_uploads, delete, dry_run, doing_init=False):
	'''Download files from production to your working copy'''
	root = util.find_root()
	config = util.config()

	if not yes and not dry_run:
		click.confirm('Downloading files from production may overwrite '
			+ 'your local copy. Continue?',
			abort=True
		)

	delete = ['--delete'] if delete else []

	if dry_run:
		return ctx.invoke(diff, path=path, reverse=True)

	util.heading('Downloading production data')
	util.item('Downloading application files from production')

	returncode, stdout, stderr = util.rsync(
		args=['-rtl'] + delete,
		source='root@%s:%s' % (config['hostname'], util.remote_path('/public/')),
		destination='%s/' % root,
		extend_filters=_get_extend_filters(path)
	)

	if returncode != 0:
		raise util.SailException('An error occurred during download. Please try again.')

	if with_uploads:
		util.item('Downloading wp-content/uploads')

		# Download uploads from production
		returncode, stdout, stderr = util.rsync(
			args=['-rtl'] + delete,
			source='root@%s:%s' % (config['hostname'], util.remote_path('/uploads/')),
			destination='%s/wp-content/uploads/' % root,
			default_filters=False,
			extend_filters=_get_extend_filters(path, 'wp-content/uploads')
		)

		if returncode != 0:
			raise util.SailException('An error occurred during download. Please try again.')

	util.item('All download tasks completed')

	# Don't show success during provision/init.
	if not doing_init:
		util.success('Files download completed')

@cli.command()
@click.argument('path', nargs=-1, required=False)
@click.option('--reverse', is_flag=True, help='Reverse the file comparison direction (production to local)')
@click.option('--raw', is_flag=True, help='Show the list of changes in raw format')
def diff(path, reverse, raw):
	'''Show file changes between your local copy and production.'''
	root = util.find_root()
	config = util.config()

	if raw:
		util.loader(suspend=True)
	else:
		util.heading('Comparing files')

	destination = 'root@%s:%s' % (config['hostname'], util.remote_path('/public/'))
	source = '%s/' % root

	if reverse:
		destination, source = source, destination

	filters = _get_deployignore_filters()
	filters += _get_extend_filters(path)
	files = _diff(source, destination, filters)
	empty = True

	colors = {'created': 'green', 'deleted': 'red', 'updated': 'yellow'}
	labels = {'created': 'New', 'deleted': 'Delete', 'updated': 'Update'}
	for op in ['created', 'updated', 'deleted']:
		for filename in files[op]:
			if empty:
				empty = False
				click.echo()

			if raw:
				click.echo(filename)
			else:
				click.secho('  %s: %s' % (labels[op], filename), fg=colors[op])

	# TODO: Compare uploads if requested --with-uploads

	if empty and not raw:
		util.item('No changes')

	if not raw:
		click.echo()

def _diff(source, destination, extend_filters=[]):
	'''Compare application files between environments'''
	root = util.find_root()

	if source == destination:
		raise util.SailException('Can not compare apples to apples')

	args = ['-rlci', '--delete', '--dry-run']
	returncode, stdout, stderr = util.rsync(args, source, destination, extend_filters=extend_filters)

	if returncode != 0:
		raise util.SailException('An error occurred in rsync. Please try again.')

	files = {
		'created': [],
		'deleted': [],
		'updated': [],
	}

	if util.debug():
		util.dlog(stdout)

	for line in stdout.splitlines():
		change = line[:11]
		path = line[12:]

		if len(change) < 11:
			continue

		# See --itemize-changes https://linux.die.net/man/1/rsync
		# Deleted
		if change.strip() == '*deleting':
			files['deleted'].append(path)
			continue

		# Newly created file.
		if change == '<f+++++++++' or change == '>f+++++++++':
			files['created'].append(path)
			continue

		# Newly created directory.
		if change == 'cd+++++++++':
			files['created'].append(path)
			continue

		# Checksum different.
		if change[:3] == '<fc' or change[:3] == '>fc':
			files['updated'].append(path)
			continue

		# Permission change
		if change[5] == 'p':
			files['updated'].append(path)
			continue

	return files



from sail import cli, util, domains, provision, blueprints, cron, __version__

import sail
import json
import subprocess, os
import click
import shutil
import digitalocean
import pathlib, hashlib, glob
import secrets, string

@cli.command()
@click.option('--yes', '-y', is_flag=True, help='Force yes on AYS prompts')
@click.option('--skip-snapshot', is_flag=True, help='Do not create a snapshot before rebuild')
@click.pass_context
def rebuild(ctx, yes, skip_snapshot):
	'''Rebuild an application environment.'''
	root = util.find_root()
	config = util.config()

	if not yes:
		click.confirm('Rebuilding will wipe the server and re-provision it from scratch. Continue?', abort=True)

	if config['namespace'] != 'default':
		raise util.SailException('Rebuild requires the default namespace. Current namespace: %s' % config['namespace'])

	c = util.connection()

	namespaces = []
	try:
		remote_config = json.loads(c.run('cat /etc/sail/config.json').stdout)
		namespaces = remote_config.get('namespaces', [])
	except:
		pass

	if len(namespaces) > 1 and not yes:
		click.confirm('This environment has multiple namespaces: %s. Wipe them all?' % ', '.join(namespaces), abort=True)

	util.heading('Rebuilding environment')

	util.item('Getting droplet data')
	try:
		droplet = digitalocean.Droplet(token=config['provider_token'], id=config['droplet_id'])
		droplet.load()
	except:
		raise util.SailException('Could not get droplet data')

	def wait_for_ssh():
		try:
			c.run('hostname')
			return True
		except Exception as e:
			util.dlog(repr(e))
			return False

	if not skip_snapshot:
		util.item('Creating a snapshot')
		action = droplet.take_snapshot(
			snapshot_name='sail-rebuild-%s' % hashlib.sha256(os.urandom(32)).hexdigest()[:8],
			return_dict=False,
			power_off=True,
		)

		if not action.wait(update_every_seconds=10, repeat=60):
			droplet.power_on() # Attempt to power-on the droplet before exit
			raise util.SailException('Could not complete snapshot action. Status: %s' % action.status)

		util.item('Waiting for power on')
		action = droplet.power_on(return_dict=False)

		if not action.wait(update_every_seconds=1, repeat=60):
			raise util.SailException('Could not power on the droplet. Status: %s' % action.status)

		util.item('Waiting for SSH')
		util.wait(wait_for_ssh, timeout=120, interval=10)

	util.item('Rebuilding droplet')

	try:
		action = droplet.rebuild(image_id=sail.DEFAULT_IMAGE, return_dict=False)
	except:
		raise util.SailException('Could not schedule rebuild action')

	util.item('Waiting for rebuild to complete')
	if not action.wait(update_every_seconds=1, repeat=60):
		raise util.SailException('Could not complete rebuild action. Status: %s' % action.status)

	util.item('Deleting .sail/known_hosts')
	try:
		pathlib.Path('%s/.sail/known_hosts' % root).unlink()
	except:
		util.item('Could not delete .sail/known_hosts')

	# Claim new application ID
	response = util.request('/init/', json={
		'email': config['email'],
		'version': __version__,
	}, anon=True)

	app_id = response['app_id']
	secret = response['secret']
	hostname = response['hostname']

	util.item('Claimed application id: %s' % app_id)

	util.item('Updating .sail/config.json')
	updated_config = {
		'app_id': app_id,
		'secret': secret,
		'hostname': hostname,
		'provider_token': config['provider_token'],
		'email': config['email'],
		'domains': [],
		'profile_key': ''.join(secrets.choice(string.ascii_letters + string.digits) for i in range(32)),
		'namespace': config['namespace'],
		'version': __version__,

		'ip': config['ip'],
		'droplet_id': config['droplet_id'],
		'key_id': config['key_id'],
	}

	config = updated_config
	util.update_config(config)

	util.item('Requesting DNS record')
	response = util.request('/ip/', json={
		'ip': config['ip'],
	})

	# With empty known_hosts
	c = util.connection()
	util.item('Waiting for SSH')
	util.wait(wait_for_ssh, timeout=120, interval=10)

	util.item('Writing server keys to .sail/known_hosts')
	with open('%s/.sail/known_hosts' % root, 'w+') as f:
		r = subprocess.run(['ssh-keyscan', '-t', 'rsa,ecdsa', '-H', config['ip'],
			config['hostname']], stdout=f, stderr=subprocess.DEVNULL)

	# With new known_hosts
	c = util.connection()
	provision._configure()

	passwords = {}
	for key in ['mysql', 'wp']:
		passwords[key] = ''.join(secrets.choice(string.ascii_letters
			+ string.digits) for i in range(48))

	# Run the installs
	provision._install(passwords)

	# Add the default WP cron schedule.
	ctx.invoke(cron.add, schedule='*/5', command=('wp cron event run --due-now',), quiet=True)

	# Run a default blueprint
	try:
		ctx.invoke(blueprints.blueprint, path=['default.yaml'], doing_init=True)
	except:
		util.item('Could not apply blueprint')

	# Create a local empty wp-contents/upload directory
	content_dir = '%s/wp-content/uploads' % root
	if not os.path.exists(content_dir):
		os.mkdir(content_dir)

	# Automatically enable premium if it's been set up
	premium_license = util.get_sail_default('premium')
	premium_email = util.get_sail_default('email')
	if premium_license and premium_email:
		try:
			ctx.invoke(sail.premium.enable, doing_init=True)
		except:
			util.item('Could not initialize Premium')

	util.heading('Rebuild successful')
	provision._success(passwords['wp'])



from sail import cli, util

import os, subprocess
import click
import hashlib
import pathlib
import json, re
import secrets, string

from datetime import datetime

@cli.group()
def db():
	'''Import and export MySQL databases, or spawn an interactive shell'''
	pass

@db.command()
def cli():
	'''Open an interactive MySQL shell on the production host'''
	root = util.find_root()
	config = util.config()

	os.execlp('ssh', 'ssh', '-t',
		'-i', '%s/.sail/ssh.key' % root,
		'-o', 'UserKnownHostsFile="%s/.sail/known_hosts"' % root,
		'-o', 'IdentitiesOnly=yes',
		'-o', 'IdentityFile="%s/.sail/ssh.key"' % root,
		'root@%s' % config['hostname'],
		'sudo -u www-data wp --path=%s db cli' % util.remote_path('/public')
	)

@db.command(name='import')
@click.argument('path', nargs=1, required=True)
@click.option('--partial', is_flag=True, help='Do not wipe production database and perform a partial import')
def import_cmd(path, partial):
	'''Import a local .sql or .sql.gz file to the production MySQL database'''
	root = util.find_root()
	config = util.config()
	c = util.connection()
	remote_path = util.remote_path()
	namespace = config['namespace']

	path = pathlib.Path(path).resolve()
	if not path.exists():
		raise util.SailException('File does not exist')

	if not path.name.endswith('.sql') and not path.name.endswith('.sql.gz'):
		raise util.SailException('This does not look like a .sql or .sql.gz file')

	temp_filename = '%s.%s' % (hashlib.sha256(os.urandom(32)).hexdigest()[:8], path.name)
	is_gz = path.name.endswith('.sql.gz')

	util.heading('Importing WordPress database')
	util.item('Uploading database file to production')

	args = ['-t']
	source = path
	destination = 'root@%s:%s/%s' % (config['hostname'], remote_path, temp_filename)
	returncode, stdout, stderr = util.rsync(args, source, destination, default_filters=False)

	if returncode != 0:
		raise util.SailException('An error occurred in rsync. Please try again.')

	cat_bin = 'zcat' if is_gz else 'cat'
	temp_name = 'import_%s' % hashlib.sha256(os.urandom(32)).hexdigest()[:8]

	try:
		# A partial import, no temp table, no replacements
		# Run as is on production db.
		if partial:
			util.item('Importing database into MySQL')
			c.run(f'{cat_bin} {remote_path}/{temp_filename} | mysql -uroot "wordpress_{namespace}"')

		# Full import. Try to clean it up as much as possible.
		else:
			util.item('Creating temporary database')
			c.run(f'mysql -uroot -e "CREATE DATABASE \\`{temp_name}\\`;"')

			util.item('Importing into temporary database')
			c.run(f'{cat_bin} {remote_path}/{temp_filename} | mysql -uroot "{temp_name}"')

			# Fetch all tables and determine table prefix
			tables = c.run(f'mysql -uroot "{temp_name}" --skip-column-names -e "SHOW TABLES;"').stdout.splitlines()
			core_tables = ['commentmeta', 'comments', 'links', 'options', 'postmeta',
				'posts', 'term_relationships', 'term_taxonomy', 'termmeta', 'terms',
				'usermeta', 'users'
			]

			prefixes = []

			for table in tables:
				prefix = re.search(r'^(.+?)(?:\d+_)?(?:%s)$' % '|'.join(core_tables), table)
				if prefix:
					prefixes.append(prefix.group(1))

			# Convert to set and make unique
			prefixes = set(prefixes)
			if len(prefixes) == 1:
				prefix = prefixes.pop()
				util.item(f'Determined table prefix: {prefix}')
			else:
				prefix = None

			clean_tables = []

			# Rename
			if prefix and prefix != 'wp_':
				for table in tables:
					if table[:len(prefix)] != prefix:
						clean_tables.append(table)
						continue

					table = table[len(prefix):]
					util.item(f'Renaming {prefix}{table} to wp_{table}')
					c.run(f'mysql -uroot "{temp_name}" -e "RENAME TABLE \\`{prefix}{table}\\` TO \\`wp_{table}\\`;"')
					clean_tables.append(f'wp_{table}')

				tables = clean_tables

				util.item(f'Updating prefix in wp_options, wp_usermeta')
				meta_keys = ['capabilities', 'user_level']
				option_names = ['user_roles']

				for meta_key in meta_keys:
					c.run(f'mysql -uroot "{temp_name}" -e "UPDATE \\`wp_usermeta\\` SET meta_key = \'wp_{meta_key}\' WHERE meta_key = \'{prefix}{meta_key}\';"')

				for option_name in option_names:
					c.run(f'mysql -uroot "{temp_name}" -e "UPDATE \\`wp_options\\` SET option_name = \'wp_{option_name}\' WHERE option_name = \'{prefix}{option_name}\';"')

			util.item('Dropping live database, moving temporary to live')
			c.run(f'mysql -uroot -e "DROP DATABASE \\`wordpress_{namespace}\\`; CREATE DATABASE \\`wordpress_{namespace}\\`;"')
			for table in tables:
				c.run(f'mysql -uroot -e "RENAME TABLE \\`{temp_name}\\`.\\`{table}\\` TO \\`wordpress_{namespace}\\`.\\`{table}\\`;"')

			util.item('Dropping temporary database')
			c.run(f'mysql -uroot -e "DROP DATABASE \\`{temp_name}\\`;"')

	except Exception as e:
		util.dlog(str(e))
		c.run(f'mysql -uroot -e "DROP DATABASE \\`{temp_name}\\`;"', warn=True)
		raise util.SailException('An error occurred in SSH. Please try again.')

	util.item('Cleaning up production')

	try:
		c.run(f'rm {remote_path}/{temp_filename}')
	except:
		raise util.SailException('An error occurred in SSH. Please try again.')

	util.success('Database imported')

@db.command()
@click.option('--json', 'as_json', is_flag=True, help='Output in JSON format')
def export(as_json):
	'''Export the production database to a local .sql.gz file'''
	root = util.find_root()
	config = util.config()
	c = util.connection()
	remote_path = util.remote_path()

	if as_json:
		util.loader(suspend=True)

	backups_dir = pathlib.Path(root + '/.backups')
	backups_dir.mkdir(parents=True, exist_ok=True)
	filename = datetime.now().strftime('%Y-%m-%d-%H%M%S.sql.gz')

	if not as_json:
		util.heading('Exporting WordPress database')

	try:
		c.run('mysqldump --quick --single-transaction --default-character-set=utf8mb4 -uroot "wordpress_%s" | gzip -c9 > %s/%s' % (config['namespace'], remote_path, filename))
	except:
		raise util.SailException('An error occurred in SSH. Please try again.')

	if not as_json:
		util.item('Export completed, downloading')

	args = ['-t']
	source = 'root@%s:%s/%s' % (config['hostname'], remote_path, filename)
	destination = '%s/%s' % (backups_dir, filename)
	returncode, stdout, stderr = util.rsync(args, source, destination, default_filters=False)

	if returncode != 0:
		raise util.SailException('An error occurred in rsync. Please try again.')

	if not as_json:
		util.item('Cleaning up production')

	try:
		c.run('rm %s/%s' % (remote_path, filename))
	except:
		raise util.SailException('An error occurred in SSH. Please try again.')

	if not as_json:
		util.success('Database export saved to .backups/%s' % filename)
	else:
		click.echo(json.dumps(destination))

@db.command()
def reset_password():
	'''Reset the WordPress database password and update wp-config.php'''
	root = util.find_root()
	config = util.config()

	util.heading('Resetting database password')
	util.item('Generating new password')
	password = ''.join(secrets.choice(string.ascii_letters + string.digits) for i in range(48))
	c = util.connection()

	util.item('Updating database password')
	c.run('mysql -e "ALTER USER \\`wordpress_%s\\`@localhost IDENTIFIED BY \'%s\'"' % (config['namespace'], password))

	util.item('Updating DB_PASSWORD in wp-config.php')
	wp = 'sudo -u www-data wp --path=%s --skip-themes --skip-plugins ' % util.remote_path('public')
	c.run(wp + util.join(['config', 'set', 'DB_PASSWORD', password]))

	util.item('Verifying DB_USER/DB_NAME')
	for name in ['DB_USER', 'DB_NAME']:
		value = c.run(wp + util.join(['config', 'get', name])).stdout.strip()
		expected = 'wordpress_%s' % config['namespace']
		if value != expected:
			util.item(f'Updating {name} in wp-config.php')
			c.run(wp + util.join(['config', 'set', name, expected]))

	util.success('Database password reset. Sync with: sail download wp-config.php')



from sail import cli, util

import click
import re, os
import json, time

from datetime import datetime, timedelta
from math import floor

@cli.group(invoke_without_command=True)
@click.pass_context
def monitor(ctx):
	'''Monitor your WordPress application uptime and health'''
	# Default subcommand for back-compat
	if not ctx.invoked_subcommand:
		return ctx.forward(status)

@monitor.command()
@click.option('--json', 'as_json', is_flag=True, help='Output results as a JSON object')
def status(as_json):
	'''Show the uptime and health of your application'''
	status = util.request('/premium/monitor/status/')

	if as_json:
		click.echo(json.dumps(status))
		return

	util.label_width(11)
	click.echo()

	# Status
	duration = status['reason']['duration']
	if duration < 0:
		duration = 0

	duration = format(str(timedelta(seconds=duration)))
	st_code = status['reason']['code']

	label = util.label('Status:')
	label_r = util.label('Reason:')

	if status['status'] == 'up':
		st_string = click.style('  UP  ', bg='green')
		click.echo(f'{label} {st_string} HTTP/{st_code} for {duration}')
	elif status['status'] == 'down':
		st_string = click.style(' DOWN ', bg='red')
		st_reason_description = click.style(f'HTTP/{st_code} ' + status['reason']['description'], fg='red')
		click.echo(f'{label} {st_string} for {duration}')
		click.echo(f'{label_r} {st_reason_description}')
	else:
		st_string = click.style(' UNKNOWN ', bg='bright_black')
		click.echo(f'{label} {st_string} for {duration}')

	def _uptime_color(v):
		if v == 0:
			return 'bright_black'
		if v >= 99.95:
			return 'white'
		if v >= 99.00:
			return 'yellow'
		return 'red'

	# Uptime
	intervals = ['24h', '7d', '30d']
	uptime = {}
	for i in intervals:
		v = floor(status['uptime'][i] * 100) / 100
		uptime[i] = click.style('{0:.2f}%'.format(v), fg=_uptime_color(v))

	label = util.label('Uptime:')
	uptime_24h, uptime_7d, uptime_30d = uptime['24h'], uptime['7d'], uptime['30d']
	click.echo(f'{label} {uptime_24h} in 24h, {uptime_7d} in 7d, {uptime_30d} in 30d')

	def _get_response_time_color(v):
		if v <= 0:
			return 'bright_black'
		if v <= 500:
			return 'white'
		if v < 1000:
			return 'yellow'
		return 'red'

	# Response Time
	response_time = status.get('response_time')
	response_time = click.style('{:,} ms'.format(response_time), fg=_get_response_time_color(response_time))
	label = util.label('Avg. Resp:')
	click.echo(f'{label} {response_time}')

	# Alerts/contacts
	state = click.style(status['alerts']['state'].title(), fg=_status_color(status['alerts']['status']))
	contacts = status['alerts']['contacts']
	unverified = status['alerts']['unverified']
	label = util.label('Alerts:')

	if unverified > 0:
		click.echo(f'{label} {state} ({contacts} contacts, {unverified} unverified)')
	else:
		click.echo(f'{label} {state} ({contacts} contacts)')

	# Render health-check results
	click.echo()
	_render_health(status)

	click.echo()

def _status_color(s):
	if s == 'ok':
		return 'white'
	elif s == 'warn':
		return 'yellow'
	elif s == 'critical':
		return 'red'
	else:
		return 'bright_black'

def _render_health(status):
	timestamp = status.get('health_timestamp', None)
	if not timestamp:
		for label in ['Core:', 'Plugins:', 'Themes:', 'Server:', 'Disk:', 'Updated:']:
			label = util.label(label)
			value = click.style('pending', fg=_status_color('unknown'))
			click.echo(f'{label} {value}')

		return

	# Core
	core_version = status['core']['version']
	latest_core_version = status['core']['latest']
	core_version_s = click.style(core_version, fg=_status_color(status['core']['status']))
	label = util.label('Core:')

	if core_version == latest_core_version:
		click.echo(f'{label} {core_version_s} (latest)')
	else:
		click.echo(f'{label} {core_version_s} (latest: {latest_core_version})')

	# Plugins
	total = status['plugins']['total']
	updates = status['plugins']['updates']
	uptodate = total - updates
	output = click.style(f'{uptodate}/{total} plugins', fg=_status_color(status['plugins']['status']))
	label = util.label('Plugins:')
	click.echo(f'{label} {output} up to date')

	# Themes
	total = status['themes']['total']
	updates = status['themes']['updates']
	uptodate = total - updates
	output = click.style(f'{uptodate}/{total} themes', fg=_status_color(status['themes']['status']))
	label = util.label('Themes:')
	click.echo(f'{label} {output} up to date')

	# Server
	updates = status['packages']['updates']
	updates = click.style(f'{updates} packages', fg=_status_color(status['packages']['status']))
	label = util.label('Server:')
	click.echo(f'{label} {updates} with updates')

	used = status['disk']['used']
	total = status['disk']['total']
	percent = used / total

	percent = click.style(f'{percent:.0%} used', fg=_status_color(status['disk']['status']))
	label = util.label('Disk:')

	used = util.sizeof_fmt(used)
	total = util.sizeof_fmt(total)

	click.echo(f'{label} {percent}, {used} of {total}')

	updated = datetime.fromtimestamp(timestamp)
	label = util.label('Updated:')
	click.echo(f'{label} {updated}')

@monitor.command()
@click.argument('minutes', nargs=1, required=True)
def snooze(minutes):
	'''Snooze alerts for a period of time'''
	timestamp = int(time.time())
	map = {'s': 1, 'm': 60, 'h': 60*60, 'd': 60*60*24}
	search = re.search(r'^(\d+)(s|h|m|d)?$', minutes)

	if not search:
		raise util.SailException('Invalid snooze format.')

	value = int(search.group(1))
	unit = search.group(2)
	if not unit:
		unit = 'm'

	timestamp += value * map[unit]

	click.echo()
	click.echo('Snoozing monitor')
	click.secho('  Requesting remote monitor to snooze', fg='bright_black')

	util.request('/premium/monitor/snooze/', method='POST', json={
		'timestamp': timestamp,
	})

	click.secho('  Snooze acknowledged', fg='bright_black')

	click.echo()
	util.success('Monitor snoozed')
	click.echo()

@monitor.command()
def unsnooze():
	'''Unsnooze monitoring alerts'''

	click.echo()
	click.echo('Unsnoozing monitor')
	click.secho('  Requesting remote monitor to unsnooze', fg='bright_black')

	util.request('/premium/monitor/snooze/', method='DELETE')

	click.secho('  Unsnooze acknowledged', fg='bright_black')

	click.echo()
	util.success('Monitor unsnoozed')
	click.echo()


@monitor.command()
@click.option('--quiet', '-q', is_flag=True, help='Do not output progress')
def enable(quiet):
	'''Enable monitoring for this application'''

	if not quiet:
		util.heading('Enabling monitoring')
		util.item('Requesting remote monitor enable')

	util.request('/premium/monitor/enable/', method='POST')

	if not quiet:
		util.item('Request acknowledged')
		util.success('Monitoring enabled. Status: sail monitor status')

	return True

@monitor.command()
@click.option('--yes', '-y', is_flag=True, help='Force yes on the AYS message')
@click.option('--quiet', '-q', is_flag=True, help='Do not output progress')
def disable(yes, quiet):
	'''Disable monitoring for this application'''
	if not yes:
		click.confirm('All monitoring and uptime data will be deleted. Are you sure?', abort=True)

	if not quiet:
		util.heading('Disabling monitoring')
		util.item('Requesting remote monitor disable')

	util.request('/premium/monitor/disable/', method='POST')

	if not quiet:
		util.item('Request acknowledged')
		util.success('Monitoring disabled.')

@monitor.group(invoke_without_command=True)
@click.pass_context
def contact(ctx):
	'''Manage contacts for monitoring alerts'''
	if not ctx.invoked_subcommand:
		return ctx.forward(list)

@contact.command()
def list():
	'''List existing contacts for monitoring alerts'''
	contacts = util.request('/premium/monitor/contacts/')

	if len(contacts) < 1:
		click.echo()
		util.failure('No contacts found')
		click.echo('Add contacts using: sail monitor contact add <subject>')
		click.echo()
		return

	util.label_width(9)
	has_pending = False

	click.echo()
	for contact in contacts:
		label = util.label('Subject:')
		subject = contact['subject']
		click.echo(f'{label} {subject}')

		label = util.label('Added:')
		added = datetime.fromtimestamp(contact['created'])
		click.echo(f'{label} {added}')

		label = util.label('Status:')
		status = contact['status']

		if status != 'ready':
			has_pending = True
			status = click.style(status, fg='red')

		click.echo(f'{label} {status}')
		click.echo()

	if has_pending:
		click.echo('Some contacts are pending verification!')
		click.echo('Verify with: sail monitor contact verify <subject> <code>')
		click.echo()

@contact.command()
@click.argument('subject', nargs=1, required=True)
def add(subject):
	'''Add a new contact for monitoring alerts'''
	request = util.request('/premium/monitor/contacts/', method='POST', json={
		'subject': subject,
	})

	if request['status'] == 'pending':
		click.echo()
		util.success('Contact added, pending verification', padding=False)
		click.echo('Verify using: sail monitor contact verify %s <code>' % subject)
		click.echo()
	elif request['status'] == 'ready':
		util.success('Contact added successfully, pre-verified', padding=False)
	else:
		util.failure('Could not add this contact')

@contact.command()
@click.argument('subject', nargs=1, required=True)
def delete(subject):
	'''Delete a monitoring contact'''
	request = util.request('/premium/monitor/contacts/', method='Delete', json={
		'subject': subject,
	})

	util.success('Contact deleted successfully', padding=False)

@contact.command()
@click.argument('subject', nargs=1, required=True)
@click.argument('code', nargs=1, required=True)
def verify(subject, code):
	'''Verify a monitoring contact'''
	request = util.request('/premium/monitor/contacts/verify/', method='POST', json={
		'subject': subject,
		'code': code,
	})

	util.success('Contact verified successfully', padding=False)



from sail import cli, util

import click, uuid, io, json
import shlex, hashlib

@cli.group()
def cron():
	'''Add, delete, view and execute system cron jobs'''
	pass

@cron.command()
@click.option('--json', 'as_json', is_flag=True, help='Output in JSON format')
def list(as_json):
	'''Get a list of active cron jobs'''
	config = util.config()

	cron_data = config.get('cron', {})

	if as_json:
		click.echo(json.dumps(cron_data))
		return

	if len(cron_data) < 1:
		click.echo('No cron entries found')
		return

	util.label_width(10)

	click.echo()
	for id, entry in cron_data.items():
		label = util.label('Id:')
		click.echo(f'{label} {id}')

		label = util.label('User:')
		user = ('www-data' if not entry['root'] else 'root')
		click.echo(f'{label} {user}')

		label = util.label('Schedule:')
		schedule = entry['schedule']
		click.echo(f'{label} {schedule}')

		label = util.label('Command:')
		command = entry['command']
		click.echo(f'{label} {command}')
		click.echo()

@cron.command()
@click.argument('id', nargs=1)
def delete(id):
	'''Delete a cron job by its id'''
	config = util.config()
	c = util.connection()

	cron_data = config.get('cron', {})

	if id not in cron_data:
		raise util.SailException('Could not find cron entry by id.')

	del cron_data[id]
	config['cron'] = cron_data

	util.heading('Deleting cron job')
	util.item('Updating .sail/config.json')
	util.update_config(config)

	util.item('Generating /etc/cron.d entry')
	_generate_cron()

	util.success('Cron job delete successfully')

@cron.command()
@click.argument('schedule', nargs=1)
@click.argument('command', nargs=-1)
@click.option('--root', is_flag=True, help='Run command as root')
@click.option('--quiet', '-q', is_flag=True, help='Suppress output')
def add(schedule, command, root, quiet):
	'''Add a new system cron job'''
	config = util.config()
	c = util.connection()

	if len(command) < 1:
		raise util.SailException('Invalid cron command.')

	command = ' '.join(command)

	schedules = {
		'hourly': '0 * * * *',
		'twicedaily': '0 */12 * * *',
		'daily': '0 0 * * *',
		'weekly': '0 0 * * 0',
	}

	if schedules.get(schedule):
		schedule = schedules.get(schedule)

	# Verify schedule
	parts = schedule.split(' ')
	if len(parts) > 5:
		raise util.SailException('Invalid schedule.')

	# Extend short-hand */1
	if len(parts) < 5:
		parts.extend(['*'] * (5 - len(parts)))

	schedule = ' '.join(parts)

	cron_data = config.get('cron', {})

	id = str(uuid.uuid4())
	if cron_data.get(id):
		raise util.SailException('UUID collision, try again.')

	cron_data[id] = {
		'root': root,
		'schedule': schedule,
		'command': command,
	}

	config['cron'] = cron_data

	if not quiet:
		util.heading('Adding a cron job')
		util.item('Updating .sail/config.json')

	util.update_config(config)

	if not quiet:
		util.item('Generating /etc/cron.d entry')

	_generate_cron()

	if not quiet:
		util.success('Cron job added successfully')

def _generate_cron():
	config = util.config()
	c = util.connection()

	cron_data = config.get('cron', {})

	c.run('rm /etc/cron.d/sail-%s' % config['namespace'], warn=True) # back-compat
	filename = 'sail-%s-%s' % (config['namespace'].replace('.', '_'), hashlib.md5(config['namespace'].encode('utf8')).hexdigest()[:8])

	if len(cron_data) < 1:
		c.run(f'rm /etc/cron.d/{filename}', warn=True)
		return

	contents = []
	contents.append('SHELL=/bin/sh')
	contents.append('PATH=/usr/local/sbin:/usr/local/bin:/sbin:/bin:/usr/sbin:/usr/bin')

	for id, entry in cron_data.items():
		schedule = entry['schedule']
		user = 'www-data' if not entry['root'] else 'root'
		command = 'cd %s && %s' % (util.remote_path('public'), entry['command'])
		contents.append(f'{schedule} {user} {command}')

	# Cron needs an empty line before EOF.
	contents.append('')

	c.put(io.StringIO('\n'.join(contents)), f'/etc/cron.d/{filename}')
	c.run(f'chmod 0600 /etc/cron.d/{filename}')

@cron.command()
@click.argument('id', nargs=1)
def run(id):
	'''Run a system cron job by id'''
	config = util.config()
	c = util.connection()

	cron_data = config.get('cron', {})

	if id not in cron_data:
		raise util.SailException('Could not find cron entry by id.')

	entry = cron_data[id]
	user = 'root' if entry['root'] else 'www-data'

	command = shlex.join(['PATH=/usr/local/sbin:/usr/local/bin:/sbin:/bin:/usr/sbin:/usr/bin',
		'cd', util.remote_path('public')])
	command += ' && ' + entry['command']

	result = c.run(shlex.join(['sudo', '-u', user, 'env', '-i', '/bin/sh', '-c', command]))

	if result.stderr:
		click.echo(result.stderr, err=True, nl=False)

	if result.stdout:
		click.echo(result.stdout, nl=False)



from sail import cli, util

import requests, json, os, subprocess, time, io, re
import click
import tldextract
import digitalocean
import urllib, pathlib, hashlib, shutil
import shlex

from datetime import datetime

@cli.group()
def domain():
	'''Add, remove and update domains associated with your site'''
	pass

@domain.command(name='list')
@click.option('--json', 'as_json', is_flag=True, help='Get list of domains as JSON')
def listcmd(as_json):
	'''List domains associated with your site'''
	config = util.config()

	if as_json:
		click.echo(json.dumps(config['domains']))
		return

	util.heading('Domains')
	click.echo()

	for domain in config['domains']:
		flags = []
		for flag in ['internal', 'https', 'primary']:
			if domain.get(flag, None):
				flags.append(flag)

		sflags = ''
		if flags:
			sflags = ' (%s)' % ', '.join(flags)

		click.echo('  ' + domain['name'] + sflags)

	click.echo()

@domain.command()
@click.argument('domain', nargs=1)
@click.option('--force', is_flag=True, help='Make primary again, even if the domain is already set as the primary one')
@click.option('--skip-replace', is_flag=True, help='Skip running search-replace routines for home, siteurl, and other URLs')
def make_primary(domain, force, skip_replace):
	'''Set a domain as primary, update siteurl/home, search-replace all links'''
	config = util.config()

	util.heading('Updating primary domain')

	if domain not in [d['name'] for d in config['domains']]:
		raise util.SailException('Can not make primary, domain does not exist')

	domain = [d for d in config['domains'] if d['name'] == domain][0]
	if domain['primary'] and not force:
		raise util.SailException('Domain %s already set as primary. Use --force to force' % domain['name'])

	c = util.connection()
	wp = 'sudo -u www-data wp --path=%s --skip-themes --skip-plugins ' % util.remote_path('/public/')

	home = c.run(wp + 'option get home').stdout.strip()
	current = urllib.parse.urlparse(home).netloc
	proto = 'https://' if domain['https'] else 'http://'
	util.item('Current primary domain: %s' % current)

	if skip_replace:
		util.item('Skipping search-replace')
	else:
		util.item('Running search-replace')

		for sproto in ['https://', 'http://']:
			c.run(wp + util.join([
				'search-replace',
				sproto + current,
				proto + domain['name'],
				'--all-tables',
			]))

		# Flush object cache
		util.item('Flushing object cache')
		c.run(wp + 'cache flush')

	# Update config.json
	util.item('Updating .sail/config.json')
	for i, d in enumerate(config['domains']):
		config['domains'][i]['primary'] = d['name'] == domain['name']

	util.update_config(config)

	if config['namespace'] == 'default':
		util.item('Renaming droplet')
		droplet = digitalocean.Droplet(token=config['provider_token'], id=config['droplet_id'])
		droplet.rename(domain['name'])

	util.success('Primary domain updated successfully')

@domain.command()
@click.argument('domains', nargs=-1)
@click.option('--agree-tos', is_flag=True)
def make_https(domains, agree_tos):
	'''Request and install SSL certificates for domains'''
	root = util.find_root()
	config = util.config()

	if not agree_tos:
		click.echo('Let\'s Encrypt ToS: https://community.letsencrypt.org/tos')
		click.confirm('Do you agree to the Let\'s Encrypt ToS?', abort=True)

	if not domains:
		raise util.SailException('At least one domain is required')

	domains, subdomains = _parse_domains(domains)

	util.heading('Requesting and installing SSL for domains')

	for domain in domains + subdomains:
		if domain.fqdn not in [d['name'] for d in config['domains']]:
			raise util.SailException('Domain %s does not exist, please add it first' % domain.fqdn)

	groups = _get_groups(domains, subdomains)

	c = util.connection()
	_doms, _subs = _parse_domains([d['name'] for d in config['domains'] if d['internal'] != True])

	for group in groups:
		names = []
		names.extend([d.fqdn for d in _doms if d.fqdn == group])
		names.extend([s.fqdn for s in _subs if s.registered_domain == group or s.fqdn == group])

		util.item('Requesting certificate for %s' % group)
		args = ['certbot', '-n', 'certonly', '-m', config['email'], '--agree-tos',
			'--standalone', '--http-01-port', '8088', '--expand',
		]

		for name in names:
			args.append('-d')
			args.append(name)

		try:
			c.run(util.join(args))
		except Exception as e:
			util.dlog(e)
			raise util.SailException('Could not obtain SSL certificate for %s. Use --debug for more info.' % group)

		# Update .sail/config.json
		for i, d in enumerate(config['domains']):
			if d['name'] in names:
				config['domains'][i]['https'] = True

		util.update_config(config)

	util.item('Generating nginx configuration')
	_update_nginx_config()

	util.success('SSL certificates installed. Don\'t forget to: sail domain make-primary')

@domain.command()
@click.option('--skip-dns', is_flag=True, help='Do not add or update DNS records')
@click.argument('domains', nargs=-1)
def add(domains, skip_dns, quiet_success=False):
	'''Add a new domain, with DNS records pointing to your site'''
	config = util.config()

	if not domains:
		raise util.SailException('At least one domain is required')

	domains, subdomains = _parse_domains(domains)

	util.heading('Adding domains')

	# Add all domains and subs to config.json
	for domain in domains + subdomains:
		if domain.fqdn not in [d['name'] for d in config['domains']]:
			config['domains'].append({
				'name': domain.fqdn,
				'internal': False,
				'primary': False,
				'https': False,
			})

			util.update_config(config)
			util.item('Adding %s to .sail/config.json' % domain.fqdn)
		else:
			util.item('Domain %s already exists is .sail/config.json' % domain.fqdn)

	util.item('Generating nginx configuration')
	_update_nginx_config()

	# Bail early if skipping
	if skip_dns:
		util.item('Skipping updating DNS records')
		util.success('Domain name added')
		return

	manager = digitalocean.Manager(token=config['provider_token'])
	existing = manager.get_all_domains()

	# Add or update real domains first.
	for domain in domains:
		if domain.fqdn not in [d.name for d in existing]:
			do_domain = digitalocean.Domain(
				token=config['provider_token'],
				name=domain.fqdn,
				ip_address=config['ip']
			)

			try:
				do_domain.create()
				existing.append(do_domain)
				util.item('Creating DNS zone and record for %s' % domain.fqdn)
			except:
				raise util.SailException('- Could not create DNS zone for %s' % domain.fqdn)

			continue

		# DNS zone exists, try and update it
		util.item('Updating DNS records for %s' % domain.fqdn)
		do_domain = digitalocean.Domain(token=config['provider_token'], name=domain.fqdn)
		records = do_domain.get_records()
		exists = False

		for r in records:
			if r.name != '@' or r.type != 'A':
				continue

			if r.data == config['ip']:
				util.item('DNS record for %s exists and is correct' % domain.fqdn)
				exists = True
				continue

			# Delete the remaining records.
			r.destroy()
			util.item('Deleting DNS record for %s, incorrect existing record' % domain.fqdn)

		if not exists:
			util.item('Adding new DNS record for %s' % domain.fqdn)
			do_domain.create_new_domain_record(name='@', type='A', data=config['ip'])

	# Add all subdomains
	for subdomain in subdomains:
		# TODO: Add support for orphaned subdomains
		if subdomain.registered_domain not in [d.name for d in existing]:
			util.item('Skipping DNS for %s, zone not found' % subdomain.fqdn)
			continue

		do_domain = digitalocean.Domain(token=config['provider_token'], name=subdomain.registered_domain)
		records = do_domain.get_records()
		exists = False

		for r in records:
			if r.name != subdomain.subdomain or r.type not in ['A', 'CNAME']:
				continue

			if r.type == 'A' and r.data == config['ip']:
				util.item('Skipping DNS record for %s, existing record is correct' % subdomain.fqdn)
				exists = True
				continue

			# Delete the remaining records.
			r.destroy()
			util.item('Deleting DNS record for %s, incorrect exesting record' % subdomain.fqdn)

		if not exists:
			util.item('Adding new DNS record for %s' % subdomain.fqdn)
			do_domain.create_new_domain_record(name=subdomain.subdomain, type='A', data=config['ip'])

	# Can be invoked from another context, avoid blending multiple
	# success messages together.
	if not quiet_success:
		util.success('Domain name(s) added')

@domain.command()
@click.option('--skip-dns', is_flag=True, help='Do not delete DNS records')
@click.option('--zone', is_flag=True, help='Force delete a DNS zone, even if other records exist in the zone.')
@click.argument('domains', nargs=-1)
def delete(domains, skip_dns, zone):
	'''Delete a domain and all DNS records'''
	config = util.config()

	if not domains:
		raise util.SailException('At least one domain is required')

	domains, subdomains = _parse_domains(domains)

	util.heading('Deleting domains')

	# Remove all domains and subs from config.json
	for domain in domains + subdomains:
		if domain.fqdn in [d['name'] for d in config['domains']]:
			config['domains'] = [d for d in config['domains'] if d['name'] != domain.fqdn]
			util.update_config(config)
			util.item('Deleting %s from .sail/config.json' % domain.fqdn)
		else:
			util.item('Domain %s does not exist in .sail/config.json' % domain.fqdn)

	util.item('Generating nginx configuration')
	_update_nginx_config()
	# TODO: Maybe delete SSL certs for this domain

	# Bail early if skipping
	if skip_dns:
		util.item('Skipping updating DNS records')
		util.success('Domain deleted')
		return

	# Delete orphans if the entire zone is going to be deleted.
	if zone:
		for domain in domains:
			_, _subs = _parse_domains([d['name'] for d in config['domains'] if d['internal'] != True])
			for _sub in _subs:
				if _sub.registered_domain == domain.fqdn:
					util.item('Deleting orphaned subdomain %s from .sail/config.json' % _sub.fqdn)
					config['domains'] = [d for d in config['domains'] if d['name'] != _sub.fqdn]
					util.update_config(config)

	_delete_dns_records(domains, subdomains, force_delete_zones=zone)

	util.success('Domain deleted')

@domain.command()
def export():
	'''Export domain configuration and SSL certificates to a local archive.'''
	root = util.find_root()
	config = util.config()
	c = util.connection()

	rsync_args = ['-rtl']
	backups_dir = pathlib.Path(root + '/.backups')
	backups_dir.mkdir(parents=True, exist_ok=True)
	filename = datetime.now().strftime('%Y-%m-%d-%H%M%S.domains.tar.gz')
	target = pathlib.Path(backups_dir / filename)

	progress_dir = pathlib.Path(backups_dir / ('.%s.progress' % hashlib.sha256(os.urandom(32)).hexdigest()[:8]))
	progress_dir.mkdir()
	(progress_dir / 'live').mkdir()
	(progress_dir / 'renewal').mkdir()

	# Get all non-internal domains, split into registered domains and subdomains.
	# Then split into groups as this is how we register them with Certbot.
	domains = [d for d in config['domains'] if not d['internal']]
	_domains, _subdomains = _parse_domains([d['name'] for d in domains])
	groups = _get_groups(_domains, _subdomains)

	if len(domains) < 1 or len(groups) < 1:
		raise util.SailException('Could not find any domains to export.')

	util.heading('Exporting domains')

	# Available certificates.
	certs = c.run('ls /etc/letsencrypt/live/').stdout.split()

	for group in groups:
		if not group in certs:
			util.item('Skipping group without certificate: %s' % group)
			continue

		# Copy certificates and renewal configuration for group.
		util.item('Copying /etc/letsencrypt/live/%s' % group)

		source = 'root@%s:/etc/letsencrypt/live/%s' % (config['hostname'], group)
		destination = '%s/live/' % progress_dir
		returncode, stdout, stderr = util.rsync(rsync_args, source, destination, default_filters=False)

		if returncode != 0:
			shutil.rmtree(progress_dir)
			raise util.SailException('An error occurred during export. Please try again.')

		util.item('Copying /etc/letsencrypt/renewal/%s.conf' % group)

		source = 'root@%s:/etc/letsencrypt/renewal/%s.conf' % (config['hostname'], group)
		destination = '%s/renewal/' % progress_dir
		returncode, stdout, stderr = util.rsync(rsync_args, source, destination, default_filters=False)

		if returncode != 0:
			shutil.rmtree(progress_dir)
			raise util.SailException('An error occurred during export. Please try again.')

	# The account is configured in renewals so make sure to bring it over.
	util.item('Copying /etc/letsencrypts/accounts')

	source = 'root@%s:/etc/letsencrypt/accounts' % config['hostname']
	destination = '%s/' % progress_dir
	returncode, stdout, stderr = util.rsync(rsync_args, source, destination, default_filters=False)

	if returncode != 0:
		shutil.rmtree(progress_dir)
		raise util.SailException('An error occurred during export. Please try again.')

	# Dump domains configuration to a JSON file.
	util.item('Writing configuration to domains.json')
	with open('%s/domains.json' % progress_dir, 'w+') as f:
		json.dump(domains, f, indent='\t')

	util.item('Archiving and compressing files')

	p = subprocess.Popen([
		'tar', ('-cvzf' if util.debug() else '-czf'), target.resolve(), '-C', progress_dir.resolve(), '.'
	])

	while p.poll() is None:
		pass

	if p.returncode != 0:
		shutil.rmtree(progress_dir)
		raise util.SailException('An error occurred during export. Please try again.')

	shutil.rmtree(progress_dir)

	util.success('Export completed at .backups/%s' % filename)

@domain.command(name='import')
@click.argument('path', nargs=1, required=True)
@click.pass_context
def import_cmd(ctx, path):
	'''Import domains and SSL certificates from an export archive.'''
	root = util.find_root()
	config = util.config()
	c = util.connection()

	path = pathlib.Path(path).resolve()
	if not path.exists():
		raise util.SailException('File does not exist')

	if not path.name.endswith('.domains.tar.gz'):
		raise util.SailException('This does not look like a .domains.tar.gz file')

	util.heading('Importing domains and SSL certificates')

	rsync_args = ['-rtl']
	backups_dir = pathlib.Path(root + '/.backups')
	backups_dir.mkdir(parents=True, exist_ok=True)
	progress_dir = pathlib.Path(backups_dir / ('.%s.progress' % hashlib.sha256(os.urandom(32)).hexdigest()[:8]))
	progress_dir.mkdir()

	util.item('Extracting domain export files')

	p = subprocess.Popen([
		'tar', ('-xzvf' if util.debug() else '-xzf'), path.resolve(), '--directory', progress_dir.resolve()
	])

	while p.poll() is None:
		util.loader()

	if p.returncode != 0:
		shutil.rmtree(progress_dir)
		raise util.SailException('An error occurred during backup. Please try again.')

	for x in progress_dir.iterdir():
		if x.name not in ['domains.json', 'live', 'renewal', 'accounts']:
			shutil.rmtree(progress_dir)
			raise util.SailException('Unexpected file in domains export: %s' % x.name)

	util.item('Reading domains.json configuration')
	with open('%s/domains.json' % progress_dir, 'r') as f:
		domains = json.load(f)

	_domains, _subdomains = _parse_domains([d['name'] for d in domains])
	groups = _get_groups(_domains, _subdomains)
	certs = [i.name for i in (progress_dir / 'live').iterdir()]

	for group in groups:
		if not group in certs:
			util.item('Skipping group without certificate: %s' % group)
			continue

		# Copy certificates and renewal configuration for group.
		util.item('Writing /etc/letsencrypt/live/%s' % group)

		source = '%s/live/%s' % (progress_dir, group)
		destination = 'root@%s:/etc/letsencrypt/live/%s' % (config['hostname'], group)
		returncode, stdout, stderr = util.rsync(rsync_args, source, destination, default_filters=False)

		if returncode != 0:
			shutil.rmtree(progress_dir)
			raise util.SailException('An error occurred during import. Please try again.')

		util.item('Writing /etc/letsencrypt/renewal/%s.conf' % group)

		source = '%s/renewal/%s.conf' % (progress_dir, group)
		destination = 'root@%s:/etc/letsencrypt/renewal/%s.conf' % (config['hostname'], group)
		returncode, stdout, stderr = util.rsync(rsync_args, source, destination, default_filters=False)

		if returncode != 0:
			shutil.rmtree(progress_dir)
			raise util.SailException('An error occurred during import. Please try again.')

	util.item('Writing /etc/letsencrypts/accounts')
	source = '%s/accounts' % progress_dir
	destination = 'root@%s:/etc/letsencrypt/accounts' % config['hostname']
	returncode, stdout, stderr = util.rsync(rsync_args, source, destination, default_filters=False)

	if returncode != 0:
		shutil.rmtree(progress_dir)
		raise util.SailException('An error occurred during import. Please try again.')

	names = [d['name'] for d in domains]
	ctx.invoke(add, domains=names, quiet_success=True)

	util.heading('Cleaning up')

	# The add command will set https to False by default for any
	# new domains, but we may have existing SSL certificates for these.
	util.item('Setting HTTPS flags')
	https_names = [d['name'] for d in domains if d['https']]
	config = util.config()
	config_modified = False

	for i, domain in enumerate(config['domains']):
		if not domain['https'] and domain['name'] in https_names:
			config['domains'][i]['https'] = True
			config_modified = True

	if config_modified:
		util.update_config(config)

	# TODO: Invoke make_primary if current primary is internal.

	util.item('Removing temporary files')
	shutil.rmtree(progress_dir)

	util.success('Domains and SSL certificates have been imported.')

def _delete_dns_records(domains, subdomains, force_delete_zones=False):
	config = util.config()
	manager = digitalocean.Manager(token=config['provider_token'])
	existing = manager.get_all_domains()

	# Remove subdomains first
	for subdomain in subdomains:
		if subdomain.registered_domain not in [d.name for d in existing]:
			util.item('Skipping DNS for %s, zone not found' % subdomain.fqdn)
			continue

		do_domain = digitalocean.Domain(token=config['provider_token'], name=subdomain.registered_domain)
		records = do_domain.get_records()
		exists = False

		for r in records:
			if r.name != subdomain.subdomain or r.type != 'A':
				continue

			if r.data == config['ip']:
				exists = True
				util.item('Deleting A record for %s' % subdomain.fqdn)
				r.destroy()

		if not exists:
			util.item('No A records to delete for %s' % subdomain.fqdn)

	# Remove domains
	for domain in domains:
		if domain.fqdn not in [d.name for d in existing]:
			util.item('Skipping DNS for %s, zone not found' % domain.fqdn)
			continue

		try:
			do_domain = digitalocean.Domain(token=config['provider_token'], name=domain.fqdn)
			records = do_domain.get_records()
			exists = False
			delete_zone = True

			if force_delete_zones:
				util.item('Deleting DNS zone for %s, forced' % domain.fqdn)
				do_domain.destroy()
				continue

			for r in records:
				if r.name == '@' and r.type == 'A' and r.data == config['ip']:
					exists = True
					util.item('Deleting A record for %s' % domain.fqdn)
					r.destroy()
					continue

				if r.name == '@' and r.type == 'NS':
					continue

				if r.name == '@' and r.type == 'SOA':
					continue

				# There are more records, don't delete the zone
				delete_zone = False

			if not exists:
				util.item('No A records to delete for %s' % domain.fqdn)

			if delete_zone:
				do_domain.destroy()
				util.item('Deleting DNS zone for %s' % domain.fqdn)
			else:
				util.item('Skipping deleting DNS zone for %s, other records exist' % domain.fqdn)
		except:
			util.item('Could not delete DNS zone for %s' % domain.fqdn)

def _update_nginx_config():
	config = util.config()
	domains = config['domains']
	c = util.connection()

	server_names = [d['name'] for d in domains]

	certificates = []
	https_names = []

	result = c.run('certbot certificates -n', warn=True)
	matches = re.findall(r'Certificate Name:.+?Private Key Path:.+?$', result.stdout, re.MULTILINE | re.DOTALL)
	for match in matches:
		cert = {}
		cert['name'] = re.search(r'Certificate Name: (.+)$', match, re.MULTILINE).group(1).strip()
		cert['cert'] = re.search(r'Certificate Path: (.+)$', match, re.MULTILINE).group(1).strip()
		cert['key'] = re.search(r'Private Key Path: (.+)$', match, re.MULTILINE).group(1).strip()

		domains = re.search(r'Domains: (.+)$', match, re.MULTILINE).group(1).strip()
		domains = domains.split(' ')
		cert['domains'] = []

		for domain in domains:
			if domain in server_names:
				cert['domains'].append(domain)

		if len(cert['domains']) < 1:
			continue

		certificates.append(cert)
		https_names.extend(cert['domains'])

	http_names = list(set(server_names).difference(set(https_names)))

	c.put(io.StringIO(util.template('nginx.server.conf', {
		'certificates': certificates,
		'https_names': https_names,
		'http_names': http_names,
		'root': util.remote_path('/public'),
	})),'/etc/nginx/conf.d/%s.conf' % config['namespace'])

	c.run('systemctl reload nginx')

# Parse list into registered domains and subdomains
def _parse_domains(input_domains):
	subdomains = []
	domains = []

	for domain in input_domains:
		ex = tldextract.extract(domain, include_psl_private_domains=True)
		if not ex.domain or not ex.suffix:
			raise util.SailException('Bad domain: %s' % domain)

		# No internal domains
		if ex.domain in ['sailed', 'justsailed'] and ex.suffix == 'io':
			raise util.SailException('Bad domain: %s' % domain)

		if ex.subdomain:
			subdomains.append(ex)
			continue

		domains.append(ex)

	return domains, subdomains

def _get_groups(domains, subdomains):
	'''Returns a list of domains and subdomains grouped by their registered domain (if any)'''
	config = util.config()
	groups = []

	for domain in domains:
		if domain.fqdn not in groups:
			groups.append(domain.fqdn)

	for subdomain in subdomains:
		if subdomain.fqdn in groups or subdomain.registered_domain in groups:
			continue

		# Parent domain exists
		if subdomain.registered_domain in [d['name'] for d in config['domains']]:
			groups.append(subdomain.registered_domain)
		else:
			groups.append(subdomain.fqdn)

	return groups


import os

from .__version__ import __version__, __title__, __description__, __url__
from .__version__ import __author__, __author_email__, __license__, __copyright__

API_BASE = 'https://sailed.io/api/1.1'
DEFAULT_IMAGE = 'ubuntu-22-04-x64'
TEMPLATES_PATH = os.path.join(os.path.dirname(__file__), 'templates')

# Suppress blowfish warning until resolved: https://github.com/paramiko/paramiko/issues/2038
import warnings
from cryptography.utils import CryptographyDeprecationWarning
with warnings.catch_warnings():
	warnings.filterwarnings('ignore', category=CryptographyDeprecationWarning)
	import paramiko

from sail import util

import click

@click.group()
@click.version_option(__version__, '--version', '-v', message='%(version)s')
@click.option('--debug', '-d', is_flag=True, help='Enable verbose debug logging')
def cli(debug):
	util.debug(debug)

# Command groups
from sail import provision, deploy, database, backups, domains, misc, profiling
from sail import ssh, blueprints, cron, premium, rebuild, sftp

# Initialize premium modules
premium.premium_init()



from sail import cli, util

import click, pathlib, json
import os, shlex, subprocess, io, time
import requests, hashlib, base64

from datetime import datetime

@cli.group(invoke_without_command=True)
@click.option('--root', is_flag=True, help='Login as the root user')
@click.pass_context
def ssh(ctx, root):
	'''Open an SSH shell, manage SSH keys and more'''
	# Default subcommand for back-compat
	if not ctx.invoked_subcommand:
		ctx.forward(shell)

@ssh.group()
def key():
	pass

@key.command()
@click.argument('path', nargs=1)
@click.option('--label', help='Provide an optional label for this key')
def add(path, label=None, quiet=False):
	'''Add an SSH key to the production server by filename or GitHub URL'''
	config = util.config()

	if not quiet: util.heading('Adding SSH keys')

	c = util.connection()

	if path.startswith('https://github.com/'):
		if not quiet: util.item('Fetching keys from GitHub')

		r = requests.get(path)
		if not r.ok:
			raise util.SailException('Could not fetch keys from GitHub')

		keys = r.text.strip().split('\n')
		if not label:
			label = path
	else:
		path = pathlib.Path(path)
		if not path.exists() or not path.is_file():
			raise util.SailException('Provided public key file does not exist')

		with path.open('r') as f:
			keys = [f.read().strip()]

	to_add = {}

	for line in keys:
		items = line.split()
		if items[0] not in ['ecdsa-sha2-nistp256', 'ecdsa-sha2-nistp384', 'ecdsa-sha2-nistp521'
			'ssh-ed25519', 'ssh-dss', 'ssh-rsa']:
			raise util.SailException('Unsupported key type: %s' % items[0])

		try:
			fp = _fingerprint(line)
		except:
			raise util.SailException('Provided public key is invalid')

		to_add[fp] = line

	existing = _list(c)

	for fingerprint, key in to_add.items():
		if fingerprint in existing:
			if not quiet: util.item('Skipped %s (already exists)' % fingerprint)
			continue

		if not quiet: util.item('Added %s' % fingerprint)
		c.run(util.join(['echo', key]) + ' >> /root/.ssh/authorized_keys', hide=True)

		config['ssh_key_meta'][fingerprint] = {
			'created': int(time.time()),
			'label': label,
			'protected': False,
		}

	if not quiet: util.item('Updating .sail/config.json')
	util.update_config(config)

	if not quiet: util.success('Done')

def _list(c):
	config = util.config()
	keys = io.BytesIO()
	c.get('/root/.ssh/authorized_keys', keys)

	data = {}

	keys.seek(0)
	for line in keys.readlines():
		line = line.decode('utf8').strip()
		items = line.split()

		if items[0] not in ['ecdsa-sha2-nistp256', 'ecdsa-sha2-nistp384', 'ecdsa-sha2-nistp521'
			'ssh-ed25519', 'ssh-dss', 'ssh-rsa']:
			continue

		try:
			fp = _fingerprint(line)
		except:
			continue

		data[fp] = {'fingerprint': fp, 'key': line, 'label': None}

		if config['ssh_key_meta'].get(fp):
			data[fp]['label'] = config['ssh_key_meta'][fp]['label']

	return data

@key.command('list')
@click.option('--json', 'as_json', is_flag=True, help='Output in JSON format')
def listcmd(as_json):
	'''List currently authorized SSH keys on the production server'''
	config = util.config()

	if not as_json:
		util.heading('Listing SSH keys')
		util.item('Gathering remote SSH keys')

	c = util.connection()
	data = _list(c)

	if as_json:
		click.echo(json.dumps(data))
		return

	click.echo()
	util.label_width(8)

	for fingerprint, key in data.items():
		key_label = 'None'
		key_created = 'Unknown'
		if config['ssh_key_meta'].get(fingerprint):
			key_label = config['ssh_key_meta'][fingerprint]['label']
			key_created = datetime.fromtimestamp(config['ssh_key_meta'][fingerprint]['created'])

		label = util.label('Hash:')
		click.echo(f'{label} {fingerprint}')

		label = util.label('Label:')
		click.echo(f'{label} {key_label}')

		label = util.label('Added:')
		click.echo(f'{label} {key_created}')

		click.echo()

@key.command()
@click.argument('hash', nargs=1)
def delete(hash):
	'''Delete an authorized SSH key from the production server'''
	config = util.config()

	root = util.find_root()
	with open('%s/.sail/ssh.key.pub' % root, 'r') as f:
		sail_pub_key = f.read()

	util.heading('Deleting SSH key')
	util.item('Computing key hash')

	c = util.connection()
	try:
		sail_fp = _fingerprint(sail_pub_key)
	except:
		raise util.SailException('Could not compute hash of local Sail key')

	if hash.lower() == sail_fp.lower():
		raise util.SailException('This looks like the Sail SSH key, will not delete')

	util.item('Fetching existing keys')
	existing = _list(c)

	try:
		index = [k.lower() for k in existing.keys()].index(hash.lower())
		if not index:
			raise Exception()
	except:
		raise util.SailException('Could not find key with this fingerprint')

	fp = list(existing.keys())[index]
	key = existing[fp]['key']

	regex = key.replace('/', '\/')
	c.run(util.join(['sed', '-i', '/^%s/d' % regex, '/root/.ssh/authorized_keys']), hide=True)

	if config['ssh_key_meta'].get(fp):
		del config['ssh_key_meta'][fp]
		util.item('Updating .sail/config.json')
		util.update_config(config)

	util.success('Removed key: %s' % fp)

@ssh.command()
@click.option('--root', is_flag=True, help='Login as the root user')
def shell(root):
	'''Open an interactive SSH shell to the production host'''
	as_root = root
	root = util.find_root()
	config = util.config()

	command = ''
	if not as_root:
		command = 'sudo -u www-data bash -c "cd %s; bash"' % util.remote_path('/public')

	extra_args = '-vtt' if util.debug() else '-tt'

	os.execlp('ssh', 'ssh', extra_args,
		'-i', '%s/.sail/ssh.key' % root,
		'-o', 'UserKnownHostsFile="%s/.sail/known_hosts"' % root,
		'-o', 'IdentitiesOnly=yes',
		'-o', 'IdentityFile="%s/.sail/ssh.key"' % root,
		'-o', 'LogLevel=QUIET',
		'root@%s' % config['hostname'],
		command
	)

@ssh.command(context_settings=dict(ignore_unknown_options=True))
@click.argument('command', nargs=-1)
@click.option('--root', is_flag=True, help='Login as the root user')
def run(command, root):
	'''Run a command via SSH and return the results'''
	as_root = root
	root = util.find_root()
	config = util.config()

	if len(command) > 1:
		command = util.join(command)
	else:
		command = ''.join(command)

	if not as_root:
		command = util.join(['sudo', '-u', 'www-data', 'bash', '-c', command])

	extra_args = '-vtt' if util.debug() else '-tt'

	os.execlp('ssh', 'ssh', extra_args,
		'-i', '%s/.sail/ssh.key' % root,
		'-o', 'UserKnownHostsFile="%s/.sail/known_hosts"' % root,
		'-o', 'IdentitiesOnly=yes',
		'-o', 'IdentityFile="%s/.sail/ssh.key"' % root,
		'-o', 'LogLevel=QUIET',
		'root@%s' % config['hostname'],
		command
	)

def _fingerprint(line):
	key = base64.b64decode(line.strip().split()[1].encode('ascii'))
	fp_plain = hashlib.md5(key).hexdigest()
	return 'MD5:' + ':'.join(a+b for a,b in zip(fp_plain[::2], fp_plain[1::2]))
